"""V1 CKAN based API."""

# TODO update documentation of functions in GitLab
# TODO test functions to make sure they can read environment variables

import logging
import os

from envidat.utils import get_url, get_url_response, load_dotenv_if_in_debug_mode

log = logging.getLogger(__name__)

# Load config from environment variables
load_dotenv_if_in_debug_mode(".env")


def get_metadata_list(
    host: str = "https://www.envidat.ch", sort_result: bool = None
) -> list:
    """Get package/metadata list from API.

    Host url as a parameter or from environment.

    Args:
        host (str): API host url. Attempts to get from environment if omitted.
            Defaults to https://www.envidat.ch.
        sort_result (bool): Sort result alphabetically by metadata name.
            Default to None.

    Returns:
        list: List of JSON formatted packages.
    """
    if "API_HOST" in os.environ:
        log.debug("Getting API host from environment variable.")
        host = os.getenv("API_HOST")

    log.info(f"Getting package list from {host}.")
    try:
        package_names = get_url(f"{host}/api/3/action/package_list").json()
    except AttributeError as e:
        log.error(e)
        log.error("Getting package names from API failed.")
        raise AttributeError("Failed to extract package names as JSON.")

    log.debug("Extracting [result] key from JSON.")
    package_names = list(package_names["result"])

    log.info(f"Returned {len(package_names)} metadata entries from API.")

    if sort_result:
        log.debug("Sorting return alphabetically.")
        package_names = sorted(package_names, reverse=False)

    return package_names


def get_protocol_and_domain(
    protocol: str = "https", domain: str = "www.envidat.ch"
) -> tuple[str, str]:
    """Extract protocol string and domain string from API host.

    Args:
        protocol (str): API host protocol. Attempts to get from environment if omitted.
            Defaults to https
        domain (str): API host domain. Attempts to get from environment if omitted.
            Defaults to www.envidat.ch

    Returns:
        tuple (<str: protocol>, <str: domain>): Protocol and domain from API host.
    """
    if "API_HOST" in os.environ:
        host = os.getenv("API_HOST")
        protocol = host.partition("://")[0]
        domain = host.partition("://")[2]
        return protocol, domain

    return protocol, domain


def get_package(
    package_name: str,
    host: str = "https://www.envidat.ch",
    path: str = "/api/action/package_show?id=",
) -> dict:
    """Get individual package (metadata entry) as dictionary from API.

    Args:
        package_name (str): API package 'name' or 'id' value.
        host (str): API host url. Attempts to get from environment if omitted.
            Defaults to "https://www.envidat.ch"
        path (str): API host path. Attempts to get from environment if omitted.
            Defaults to "api/action/package_show?id="

    Returns:
        dict: Dictionary of package (metadata entry).
    """
    if "API_HOST" in os.environ and "API_PACKAGE_SHOW" in os.environ:
        log.debug("Getting API host and path from environment variables.")
        host = os.getenv("API_HOST")
        path = os.getenv("API_PACKAGE_SHOW")

    log.info(f"Getting package from {host}.")
    try:
        # Extract result dictionary from API call
        json_data = get_url(f"{host}{path}{package_name}").json()
        package = json_data["result"]
    except AttributeError as e:
        log.error(e)
        log.error("Getting package from API failed.")
        raise AttributeError("Failed to extract package as JSON.")

    return package


# TODO refactor this or get_package() as they have similar functionality,
#  check usage of functions in project
def get_envidat_record(
    package_name: str,
    host: str = "https://www.envidat.ch",
    path: str = "/api/action/package_show?id=",
    cookie: str | None = None,
) -> dict | None:
    """Get individual EnviDat record (metadata entry) as dictionary from API.

    Args:
        package_name (str): API package 'name' or 'id' value.
        host (str): API host url. Attempts to get from environment if omitted.
            Defaults to "https://www.envidat.ch"
        path (str): API host path. Attempts to get from environment if omitted.
            Defaults to "api/action/package_show?id="
        cookie (str | None): Cookie passed to API call in header,
                             default value is None as this argument is not
                             always used

    Returns:
        dict: Dictionary of package (metadata entry).
    """
    # Extract environment variables from config needed to call CKAN
    # If environment variables cannot be extracted then use default values
    # for host and path
    try:
        host = os.environ("API_HOST")
        path = os.environ("API_PACKAGE_SHOW")
    except KeyError as e:
        log.error(f"KeyError: {e} does not exist in environment vars")
    except AttributeError as e:
        log.error(e)

    try:
        # Extract result dictionary from API call, pass cookie if is truthy
        if cookie:
            response = get_url_response(f"{host}{path}{package_name}", cookie=cookie)
        else:
            response = get_url_response(f"{host}{path}{package_name}")

        # TODO improve error handling
        # Handle HTTPError from API call
        if response.status_code != 200:
            return {"status_code": response.status_code, "result": response.content}

        # Return package (derived from "result" key in response)
        if response:
            data = response.json()
            package = data["result"]
            return {"status_code": response.status_code, "result": package}

        # TODO handle if response is None

    except AttributeError as e:
        log.error(e)
        return {
            "status_code": 500,
            "result": "Failed to extract package as JSON from API, check logs",
        }


def get_metadata_json_with_resources(
    host: str = "https://www.envidat.ch",
    path: str = "/api/3/action/current_package_list_with_resources?limit=100000",
) -> dict:
    """Get all current package/metadata as dictionary with associated resources from
    API.

    Args:
        host (str): API host url. Attempts to get from environment if omitted.
            Defaults to https://www.envidat.ch
        path (str): API host path. Attempts to get from environment if omitted.
            Defaults to /api/3/action/current_package_list_with_resources?limit=100000

    Note:
        Limits results to 100000, otherwise returns only 10 results.

    Returns:
        dict:  Dictionary of packages, with nested resources.
    """
    if (
        "API_HOST" in os.environ
        and "API_PATH_CURRENT_PACKAGE_LIST_WITH_RESOURCES" in os.environ
    ):
        log.debug("Getting API host and path from environment variables.")
        host = os.getenv("API_HOST")
        path = os.getenv("API_PATH_CURRENT_PACKAGE_LIST_WITH_RESOURCES")

    log.info(f"Getting package list with resources from {host}.")
    try:
        package_names_with_resources = get_url(f"{host}{path}").json()
    except AttributeError as e:
        log.error(e)
        log.error("Getting package names with resources from API failed.")
        raise AttributeError("Failed to extract package names as JSON.")

    return package_names_with_resources


def get_metadata_list_with_resources(sort_result: bool = None) -> list:
    """Get all current package/metadata as list of results with associated resources.

    Args:
        sort_result (bool): Sort result alphabetically by metadata name.
            Default to None.

    Note:
        Limits results to 100000, otherwise returns only 10 results.

    Returns:
        list: List of packages, with nested resources.
    """
    # Get package/metadata as string in JSON format with associated resources from API
    package_names_with_resources = get_metadata_json_with_resources()

    # Extract results and assign them to a list
    log.debug("Extracting [result] key from JSON.")
    package_names_with_resources = list(package_names_with_resources["result"])
    log.info(f"Returned {len(package_names_with_resources)} metadata entries from API.")

    # If sort_result true sort by name key alphabetically
    if sort_result:
        log.debug("Sorting return by nested 'name' key alphabetically.")
        package_names_with_resources = sorted(
            package_names_with_resources, key=lambda x: x["name"], reverse=False
        )

    return package_names_with_resources


def get_metadata_name_doi() -> dict:
    """Get all current package/metadata names and DOIs as a dictionary.

    Note:
        Packages that do not have DOIs are assigned a default value
        of an empty string ''.

    Returns:
        dict: Dictionary of package information with names as keys
        and associated DOIs as values.
    """
    all_packages = get_metadata_list_with_resources()
    return {package.get("name"): package.get("doi", "") for package in all_packages}



"""ISO19139."""

import collections
import copy
import json
import string
from logging import getLogger
from urllib.parse import urlparse

from dateutil.parser import parse
from xmltodict import unparse

from envidat.api.v1 import get_protocol_and_domain

log = getLogger(__name__)


def convert_iso(metadata_record: dict) -> str:
    """Generate XML formatted string compatible with ISO19139 standard.

    Note:
        Converter is only valid for the metadata schema for EnviDat.

    Args:
        metadata_record (dict): Individual EnviDat metadata entry record dictionary.

    Returns:
        str: XML formatted string compatible with ISO9139 standard.

    """
    try:
        converted_dict = iso_convert_dataset(
            metadata_record
        )  # Convert package dictionary to OrderedDict
        return unparse(converted_dict, pretty=True)  # Convert OrderedDict to XML
    except ValueError as e:
        log.error(e)
        log.error("Cannot convert package to ISO19139 format.")
        raise ValueError("Failed to convert package to ISO19139 format.")


def iso_convert_dataset(dataset_dict: dict):
    """Create the ISO XML from API dictionary."""
    md_metadata_dict = collections.OrderedDict()

    # Header
    md_metadata_dict["@xmlns:gmd"] = "http://www.isotc211.org/2005/gmd"
    md_metadata_dict["@xmlns:xsi"] = "http://www.w3.org/2001/XMLSchema-instance"
    md_metadata_dict["@xmlns:gml"] = "http://www.opengis.net/gml"
    md_metadata_dict["@xmlns:gts"] = "http://www.isotc211.org/2005/gts"
    md_metadata_dict["@xmlns:gco"] = "http://www.isotc211.org/2005/gco"
    md_metadata_dict["@xmlns:geonet"] = "http://www.fao.org/geonetwork"
    md_metadata_dict["@xmlns:csw"] = "http://www.opengis.net/cat/csw/2.0.2"
    md_metadata_dict["@xmlns:srv"] = "http://www.isotc211.org/2005/srv"
    md_metadata_dict["@xmlns:gmx"] = "http://www.isotc211.org/2005/gmx"

    namespace = "http://www.isotc211.org/2005/gmd"
    schema = "http://www.isotc211.org/2005/gmd/gmd.xsd"
    md_metadata_dict["@xsi:schemaLocation"] = f"{namespace} {schema}"

    # File Identifier (O)
    identifier = dataset_dict.get("id", "")
    doi = dataset_dict.get("doi", "")
    if doi:
        identifier = "doi:" + doi.strip()
    md_metadata_dict["gmd:fileIdentifier"] = {"gco:CharacterString": identifier}

    # Metadata language (C) 3-letter from ISO 639-2/B
    iso_language = get_iso_language_code(dataset_dict.get("language", "en"))
    md_metadata_dict["gmd:language"] = {"gco:CharacterString": iso_language}

    # Dataset character set (C) : Defaulting to UTF-8
    md_metadata_dict["gmd:characterSet"] = {
        "gmd:MD_CharacterSetCode": {
            "@codeListValue": "MD_CharacterSetCode_utf8",
            "@codeList": (
                "http://www.isotc211.org/2005/resources/"
                "Codelist/gmxCodelists.xml#MD_CharacterSetCode"
            ),
            "@codeSpace": "ISOTC211/19115",
            "#text": "MD_CharacterSetCode_utf8",
        }
    }

    # Hierarchy Level (O)
    md_metadata_dict["gmd:hierarchyLevel"] = {
        "gmd:MD_ScopeCode": {
            "@codeList": (
                "http://www.isotc211.org/2005/resources"
                "/Codelist/gmxCodelists.xml#MD_ScopeCode"
            ),
            "@codeListValue": "dataset",
        }
    }

    # Point of Contact (M)
    maintainer = json.loads(dataset_dict.get("maintainer", "{}"))
    maintainer_name = ""
    if maintainer.get("given_name"):
        maintainer_name += maintainer["given_name"].strip() + " "
    maintainer_name += maintainer["name"]

    responsible_party_contact = collections.OrderedDict()

    responsible_party_contact["gmd:individualName"] = {
        "gco:CharacterString": maintainer_name
    }
    responsible_party_contact["gmd:organisationName"] = {
        "gco:CharacterString": maintainer.get("affiliation", "")
    }
    # responsible_party_contact['gmd:positionName'] =
    # {'gco:CharacterString':'', '@gco:nilReason':"missing"}

    rpc_ci_contact = collections.OrderedDict()
    # rpc_ci_contact['gmd:phone'] =
    # {'gmd:CI_Telephone':collections.OrderedDict()}
    # rpc_ci_contact['gmd:phone']['gmd:CI_Telephone']['gmd:voice'] =
    #  {'gco:CharacterString':'', '@gco:nilReason':"missing"}
    # rpc_ci_contact['gmd:phone']['gmd:CI_Telephone']['gmd:facsimile'] =
    #  {'gco:CharacterString':'', '@gco:nilReason':"missing"}

    rpc_ci_contact["gmd:address"] = {"gmd:CI_Address": collections.OrderedDict()}
    # rpc_ci_contact['gmd:address']['gmd:CI_Address']['gmd:deliveryPoint'] =
    # {'gco:CharacterString':'', '@gco:nilReason':"missing"}
    # rpc_ci_contact['gmd:address']['gmd:CI_Address']['gmd:city'] =
    # {'gco:CharacterString':'', '@gco:nilReason':"missing"}
    # rpc_ci_contact['gmd:address']['gmd:CI_Address']['gmd:administrativeArea'] =
    # {'gco:CharacterString':'', '@gco:nilReason':"missing"}
    # rpc_ci_contact['gmd:address']['gmd:CI_Address']['gmd:postalCode'] =
    # {'gco:CharacterString':'', '@gco:nilReason':"missing"}
    # rpc_ci_contact['gmd:address']['gmd:CI_Address']['gmd:country'] =
    # {'gco:CharacterString':'Switzerland'}
    rpc_ci_contact["gmd:address"]["gmd:CI_Address"][
        "gmd:electronicMailAddress"
    ] = get_or_missing(maintainer, "email")

    responsible_party_contact["gmd:contactInfo"] = {"gmd:CI_Contact": rpc_ci_contact}

    responsible_party_contact["gmd:role"] = {
        "gmd:CI_RoleCode": {
            "@codeListValue": "pointOfContact",
            "@codeList": (
                "http://www.isotc211.org/2005/resources/"
                "Codelist/gmxCodelists.xml#CI_RoleCode"
            ),
        }
    }

    md_metadata_dict["gmd:contact"] = {
        "gmd:CI_ResponsibleParty": responsible_party_contact
    }

    # Metadata Creation Date (M)
    metadata_created = parse(dataset_dict.get("metadata_created", "")).strftime(
        "%Y-%m-%dT%H:%M:%S"
    )
    md_metadata_dict["gmd:dateStamp"] = {"gco:DateTime": {"#text": metadata_created}}

    # Metadata Standard and Version (O)
    md_metadata_dict["gmd:metadataStandardName"] = {
        "gco:CharacterString": {"#text": "ISO 19115:2003/19139"}
    }
    md_metadata_dict["gmd:metadataStandardVersion"] = {
        "gco:CharacterString": {"#text": "1.0"}
    }

    # Reference System Information (O)
    md_metadata_dict["gmd:referenceSystemInfo"] = {
        "gmd:MD_ReferenceSystem": {
            "gmd:referenceSystemIdentifier": {
                "gmd:RS_Identifier": {
                    "gmd:code": {"gco:CharacterString": {"#text": "EPSG:4326"}}
                }
            }
        }
    }

    # Identification Info (mandatory subelements)
    md_data_id = collections.OrderedDict()

    # citation
    citation_dict = collections.OrderedDict()
    citation_dict["gmd:title"] = {"gco:CharacterString": dataset_dict.get("title", "")}
    citation_dict["gmd:date"] = {"gmd:CI_Date": collections.OrderedDict()}
    citation_dict["gmd:date"]["gmd:CI_Date"]["gmd:date"] = {
        "gco:Date": get_publication_date(dataset_dict)
    }
    citation_dict["gmd:date"]["gmd:CI_Date"]["gmd:dateType"] = {
        "gmd:CI_DateTypeCode": {
            "@codeListValue": "publication",
            "@codeList": (
                "http://www.isotc211.org/2005/resources/Codelist/"
                "gmxCodelists.xml#CI_DateTypeCode"
            ),
        }
    }

    presentation_form = cap_code(dataset_dict.get("resource_type", ""))
    if presentation_form:
        citation_dict["gmd:presentationForm"] = {
            "gmd:CI_PresentationFormCode": {
                "@codeListValue": presentation_form,
                "@codeList": (
                    "http://www.isotc211.org/2005/resources/"
                    "Codelist/gmxCodelists.xml#CI_PresentationFormCode"
                ),
            }
        }
    md_data_id["gmd:citation"] = {"gmd:CI_Citation": citation_dict}

    # abstract
    md_data_id["gmd:abstract"] = {
        "gco:CharacterString": dataset_dict.get("notes", "")
        .replace("\n", " ")
        .replace("\r", " ")
    }

    # Get extras_dict
    extras_dict = extras_as_dict(dataset_dict.get("extras", {}))

    # purpose (only in extras)
    purpose = get_ignore_case(extras_dict, "purpose")
    if not purpose:
        purpose = get_ignore_case(extras_dict, "CUSTOM_PURPOSE")
    if purpose:
        md_data_id["gmd:purpose"] = purpose
    # status (only in extras)
    status = get_ignore_case(extras_dict, "status")
    if status:
        md_data_id["gmd:status"] = {
            "gmd:MD_ProgressCode": {
                "@codeListValue": cap_code(status),
                "@codeList": (
                    "http://www.isotc211.org/2005/resources/"
                    "Codelist/gmxCodelists.xml#MD_ProgressCode"
                ),
            }
        }
    # Point of contact (copy)
    md_data_id["gmd:pointOfContact"] = {
        "gmd:CI_ResponsibleParty": copy.deepcopy(responsible_party_contact)
    }

    # Maintenance (only in extras)
    maintenance = get_ignore_case(extras_dict, "maintenance")
    if maintenance:
        md_data_id["gmd:resourceMaintenance"] = {
            "gmd:MD_MaintenanceInformation": {
                "gmd:maintenanceAndUpdateFrequency": {
                    "gmd:MD_MaintenanceFrequencyCode": {
                        "@codeListValue": cap_code(maintenance),
                        "@codeList": (
                            "http://www.isotc211.org/2005/resources/"
                            "Codelist/gmxCodelists.xml#MD_MaintenanceFrequencyCode"
                        ),
                    }
                }
            }
        }

    # graphic overview

    # keywords (type default to theme)
    keywords = get_keywords(dataset_dict)
    if keywords:
        keyword_type = {
            "gmd:MD_KeywordTypeCode": {
                "@codeListValue": "theme",
                "@codeList": (
                    "http://www.isotc211.org/2005/resources/"
                    "Codelist/gmxCodelists.xml#MD_KeywordTypeCode"
                ),
            }
        }
        md_data_id["gmd:descriptiveKeywords"] = {
            "gmd:MD_Keywords": {"gmd:keyword": keywords, "gmd:type": keyword_type}
        }

    # resource constraints (DRAFT)
    md_legal_constraints = collections.OrderedDict()
    access_constraints = get_ignore_case(extras_dict, "accessConstraints")
    if access_constraints:
        md_legal_constraints["gmd:accessConstraints"] = {
            "gmd:MD_RestrictionCode": {
                "@codeListValue": cap_code(access_constraints),
                "@codeList": (
                    "http://www.isotc211.org/2005/resources/"
                    "Codelist/gmxCodelists.xml#MD_RestrictionCode"
                ),
            }
        }
    use_constraints = get_ignore_case(extras_dict, "useConstraints")
    if use_constraints:
        md_legal_constraints["gmd:useConstraints"] = {
            "gmd:MD_RestrictionCode": {
                "@codeListValue": cap_code(use_constraints),
                "@codeList": (
                    "http://www.isotc211.org/2005/resources/"
                    "Codelist/gmxCodelists.xml#MD_RestrictionCode"
                ),
            }
        }
    md_legal_constraints["gmd:otherConstraints"] = get_or_missing(
        dataset_dict, "license_title"
    )

    md_data_id["gmd:resourceConstraints"] = {
        "gmd:MD_LegalConstraints": md_legal_constraints
    }

    # spatialRepresentationType
    # spatialResolution

    # language (copy)
    md_data_id["gmd:language"] = copy.deepcopy(md_metadata_dict["gmd:language"])

    # character set (copy)
    md_data_id["gmd:characterSet"] = copy.deepcopy(md_metadata_dict["gmd:characterSet"])

    # Topic category (only in extras)
    category = get_ignore_case(extras_dict, "category")
    if category:
        md_data_id["gmd:topicCategory"] = {
            "gmd:MD_TopicCategoryCode": cap_code(category)
        }

    # temporal extent
    dates = []
    try:
        dates = json.loads(dataset_dict.get("date", "[]"))
    except ValueError:
        dates = []
    gml_id_index = 0
    for date in dates:
        gml_id_index += 1
        date_id = "D" + "%03d" % (gml_id_index,)
        date_description = date.get("date_type", "")
        time_element = {}
        time_sub_element = collections.OrderedDict()
        if date.get("end_date"):
            time_sub_element["@gml:id"] = date_id
            time_sub_element["gml:beginPosition"] = date.get("date", "")
            time_sub_element["gml:endPosition"] = date.get("end_date", "")
            time_element = {"gml:TimePeriod": time_sub_element}
        else:
            time_sub_element["@gml:id"] = date_id
            time_sub_element["gml:timePosition"] = date.get("date", "")
            time_element = {"gml:TimeInstant": time_sub_element}

        time_extent = {"gmd:extent": time_element}
        md_data_id["gmd:extent"] = {"gmd:EX_Extent": collections.OrderedDict()}
        md_data_id["gmd:extent"]["gmd:EX_Extent"]["gmd:description"] = {
            "gco:CharacterString": date_description
        }
        md_data_id["gmd:extent"]["gmd:EX_Extent"]["gmd:temporalElement"] = {
            "gmd:EX_TemporalExtent": time_extent
        }

    # geographic extent
    try:
        spatial = json.loads(dataset_dict.get("spatial", "{}"))
    except ValueError:
        spatial = {}
    if spatial:
        geographic_element = collections.OrderedDict()
        if spatial.get("type") == "Point":
            gml_id_index += 1
            point_id = "P" + "%03d" % (gml_id_index,)
            coordinates = []
            for coordinate in spatial.get("coordinates", []):
                coordinates += [str(coordinate)]
            point_element = {"@gml:id": point_id, "gml:pos": " ".join(coordinates)}
            geographic_element = {
                "gmd:EX_BoundingPolygon": {"gmd:polygon": {"gml:Point": point_element}}
            }
        elif spatial.get("type") == "MultiPoint":
            gml_id_index += 1
            multi_point_id = "MP" + "%03d" % (gml_id_index,)
            multi_point_element = {"@gml:id": multi_point_id, "gml:pointMember": []}
            for coordinate_pair in spatial.get("coordinates", []):
                gml_id_index += 1
                point_id = "P" + "%03d" % (gml_id_index,)
                coordinates = " ".join(
                    [str(coordinate_pair[0]), str(coordinate_pair[1])]
                )
                point_element = {
                    "gml:Point": {"@gml:id": point_id, "gml:pos": coordinates}
                }
                multi_point_element["gml:pointMember"] += [point_element]
            geographic_element = {
                "gmd:EX_BoundingPolygon": {
                    "gmd:polygon": {"gml:MultiPoint": multi_point_element}
                }
            }
        else:
            # coordinates = spatial.get('coordinates', [])[0]
            spatial_coordinates = spatial.get("coordinates", [])
            if len(spatial_coordinates) > 0:
                coordinates = spatial_coordinates[0]

                if is_a_box(coordinates):
                    bounding_box = collections.OrderedDict()
                    bounding_box["gmd:westBoundLongitude"] = {
                        "gco:Decimal": str(min(coordinates[0][0], coordinates[2][0]))
                    }
                    bounding_box["gmd:eastBoundLongitude"] = {
                        "gco:Decimal": str(max(coordinates[0][0], coordinates[2][0]))
                    }
                    bounding_box["gmd:southBoundLatitude"] = {
                        "gco:Decimal": str(min(coordinates[0][1], coordinates[2][1]))
                    }
                    bounding_box["gmd:northBoundLatitude"] = {
                        "gco:Decimal": str(max(coordinates[0][1], coordinates[2][1]))
                    }
                    geographic_element = {"gmd:EX_GeographicBoundingBox": bounding_box}
                elif coordinates:
                    gml_id_index += 1
                    polygon_id = "PL" + "%03d" % (gml_id_index,)
                    pos_list = []
                    for coordinate_pair in coordinates:
                        coordinates = " ".join(
                            [str(coordinate_pair[0]), str(coordinate_pair[1])]
                        )
                        pos_list += [coordinates]
                        # if len(coordinate_pair) == 2:
                        #     coordinates = ' '.join(
                        #         [str(coordinate_pair[0]), str(coordinate_pair[1]
                        #     )])
                        #     pos_list += [coordinates]
                    polygon_element = {
                        "@gml:id": polygon_id,
                        "gml:interior": {"gml:LinearRing": {"gml:pos": pos_list}},
                    }
                    geographic_element = {
                        "gmd:EX_BoundingPolygon": {
                            "gmd:polygon": {"gml:Polygon": polygon_element}
                        }
                    }
                    # spatial.get(type) == 'Polygon':
        md_data_id["gmd:extent"] = {
            "gmd:EX_Extent": {"gmd:geographicElement": geographic_element}
        }

    # assign to parent
    md_metadata_dict["gmd:identificationInfo"] = {
        "gmd:MD_DataIdentification": md_data_id
    }

    # distribution info
    md_data_dist = collections.OrderedDict()

    # distribution format
    resource_formats = []
    distribution_formats = []
    for resource in dataset_dict.get("resources", []):
        resource_format = resource.get(
            "format", resource.get("mimetype", resource.get("mimetype_inner", ""))
        )
        if resource_format:
            resource_format = resource_format.upper()
            if resource_format not in resource_formats:
                resource_formats += [resource_format]
                md_format = collections.OrderedDict()
                md_format["gmd:name"] = {"gco:CharacterString": resource_format}
                md_format["gmd:version"] = {"gco:CharacterString": ""}
                distribution_formats += [{"gmd:MD_Format": md_format}]

    md_data_dist["gmd:distributionFormat"] = distribution_formats

    # assign to parent
    md_metadata_dict["gmd:distributionInfo"] = collections.OrderedDict()
    md_metadata_dict["gmd:distributionInfo"]["gmd:MD_Distribution"] = md_data_dist

    # Dataset url as information
    protocol, host = get_protocol_and_domain()
    package_name = dataset_dict.get("name", "")
    package_url = f"{protocol}://{host}/dataset/{package_name}"

    # transfer options
    online_resources = []
    online_resource_dataset = get_online_resource(
        package_url, "dataset metadata", "information"
    )
    online_resources += [online_resource_dataset]

    # loop through resources
    for resource in dataset_dict.get("resources", []):

        resource_name = resource.get("name", resource.get("id", "DATASET RESOURCE"))
        resource_id = resource.get("id", "")
        resource_url = resource.get("url", "")

        # check if restricted
        if not is_url(str(resource_url)):
            log.debug(f"resource is restricted: {resource_name}")
            resource_url = f"{package_url}/resource/{resource_id}"

        log.debug([resource_url, resource_name])
        online_resource = get_online_resource(resource_url, resource_name)
        online_resources += [online_resource]

    # assign to parent
    md_metadata_dict["gmd:distributionInfo"]["gmd:MD_Distribution"][
        "gmd:transferOptions"
    ] = {"gmd:MD_DigitalTransferOptions": {"gmd:onLine": online_resources}}

    # Root element
    iso_dict = collections.OrderedDict()
    iso_dict["gmd:MD_Metadata"] = md_metadata_dict

    return iso_dict


def extras_as_dict(extras: list) -> dict:
    """Convert API 'extras' field to simple dictionary."""
    extras_dict = {}
    for extra in extras:
        extras_dict[extra.get("key")] = extra.get("value")
    return extras_dict


def get_iso_language_code(code: str) -> str:
    """Translate to language to 3-letter code.

    http://www.loc.gov/standards/iso639-2/ISO-639-2_utf-8.txt
    """
    lookup_dict = {"en": "eng", "de": "ger", "it": "ita", "fr": "fre", "ro": "roh"}
    return lookup_dict.get(code, "eng").title()


def get_or_missing(data_dict: dict, tag, ignore_case: bool = False) -> dict:
    """TODO."""
    if ignore_case:
        if get_ignore_case(data_dict, tag):
            return {"gco:CharacterString": get_ignore_case(data_dict, tag)}
    else:
        if data_dict.get(tag):
            return {"gco:CharacterString": data_dict.get(tag)}

    return {"gco:CharacterString": "", "@gco:nilReason": "missing"}


def get_ignore_case(data_dict: dict, tag: str, ignore_blanks: bool = True) -> str:
    """TODO."""
    tag_lower = tag.lower()
    if ignore_blanks:
        tag_lower = tag_lower.replace(" ", "")
    tag_key = ""
    for key in data_dict.keys():
        key_lower = key.lower()
        if ignore_blanks:
            key_lower = key_lower.replace(" ", "")
        if key_lower == tag_lower:
            tag_key = key
            break
    return data_dict.get(tag_key)


def get_publication_date(data_dict: dict) -> str:
    """Take date of type available or the publication year.

    Returns:
        str: Publication date from dataset, in string format.
    """
    publication_date = ""
    try:
        dates = json.loads(data_dict.get("date", "[]"))
    except ValueError:
        dates = []
    for date in dates:
        if date.get("date_type") == "available":
            publication_date = parse(date.get("date")).strftime("%Y-%m-%d")
    if not publication_date:
        publication = json.loads(data_dict.get("publication", "{}"))
        publication_date = parse(publication["publication_year"] + "-12-31").strftime(
            "%Y-%m-%d"
        )
    return publication_date


def cap_code(text: str) -> str:
    """Make capcase code."""
    if text:
        text = text.strip()
        if text.find(" ") >= 0 and len(text) >= 3:
            first_word = (text.split(" ", 1)[0]).lower()
            cap_words = "".join(string.capwords(text.split(" ", 1)[1]).split(" "))
            return first_word + cap_words
        else:
            text = text[0].lower() + text[1:]
    return text


def get_keywords(data_dict: dict) -> str:
    """Extract keywords from tags."""
    keywords = []
    for tag in data_dict.get("tags", []):
        name = tag.get("display_name", "").upper()
        keywords += [{"gco:CharacterString": name}]
    return keywords


def is_a_box(coordinates: list) -> bool:
    """Check if polygon coordinates form a box."""
    if len(coordinates) == 5:
        if coordinates[0] == coordinates[4]:
            if (
                (coordinates[1] == [coordinates[0][0], coordinates[2][1]])
                and (coordinates[3] == [coordinates[2][0], coordinates[0][1]])
                and (coordinates[0][0] != coordinates[2][0])
                and (coordinates[0][1] != coordinates[2][1])
            ):
                return True
    return False


def get_online_resource(url: str, name: str, function: str = "download") -> dict:
    """Create an online resource digital transfer element."""
    protocol = url.split(":")[0]

    online_resource_dataset = {"gmd:CI_OnlineResource": collections.OrderedDict()}

    online_resource_dataset["gmd:CI_OnlineResource"]["gmd:linkage"] = {"gmd:URL": url}
    online_resource_dataset["gmd:CI_OnlineResource"]["gmd:protocol"] = {
        "gco:CharacterString": protocol.upper()
    }
    online_resource_dataset["gmd:CI_OnlineResource"]["gmd:name"] = {
        "gco:CharacterString": name.upper()
    }
    online_resource_dataset["gmd:CI_OnlineResource"]["gmd:function"] = {
        "gmd:CI_OnLineFunctionCode": {
            "@codeList": (
                "http://www.isotc211.org/2005/resources/"
                "Codelist/gmxCodelists.xml#CI_OnLineFunctionCode"
            ),
            "@codeListValue": function,
            "#text": function,
        }
    }
    return online_resource_dataset


def is_url(url_str):
    """Check if string is URL.

    Replicates functionality of CKAN method ckan.lib.helpers.is_url()

    Returns:
        bool: True if argument parses as a http, https or ftp URL,
              else returns False.
    """
    parts = urlparse(url_str)
    scheme = parts.scheme
    if scheme in ["http", "https", "ftp"]:
        return True
    return False



import json

import pytest
from moto import mock_s3

from envidat.s3.bucket import Bucket
from envidat.s3.exceptions import NoSuchCORSConfiguration


@mock_s3
def test_get_s3_resource(bucket):
    resource = Bucket.get_boto3_resource()
    assert resource, "No boto3 resource was returned"


@mock_s3
def test_get_s3_client(bucket):
    client = Bucket.get_boto3_client()
    assert client, "No boto3 client was returned"


@mock_s3
def test_bucket_create_public(bucket):
    bucket.is_public = True
    new_bucket = bucket.create()
    # Must reset class variable override for other tests
    bucket.is_public = False

    response = new_bucket.meta.client.head_bucket(Bucket="testing")
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200


@mock_s3
def test_set_public_read(bucket):
    bucket.create()

    success = bucket.set_public_read()
    assert success is True

    client = bucket.get_boto3_client()
    response = client.get_bucket_policy(
        Bucket=bucket.bucket_name,
    )
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    policy = json.loads(response["Policy"])
    assert policy["Statement"][0]["Sid"] == "PublicRead"


@mock_s3
def test_grant_full_access(bucket):
    bucket.create()

    acl = bucket.grant_user_full_access("ffffffff")
    assert acl["Grants"][1]["Grantee"]["ID"] == "ffffffff"


@mock_s3
def test_remove_full_access(bucket):
    bucket.create()

    acl = bucket.grant_user_full_access("ffffffff")
    assert acl["Grants"][1]["Grantee"]["ID"] == "ffffffff"

    acl = bucket.remove_user_full_access("user_that_does_not_exist")
    assert acl["Grants"][1]["Grantee"]["ID"] == "ffffffff"

    acl = bucket.remove_user_full_access("ffffffff")
    assert len(acl["Grants"]) == 1


@mock_s3
def test_configure_static_website(bucket):
    bucket.create()

    success = bucket.configure_static_website()
    assert success is True


@mock_s3
def test_generate_index_html(bucket):
    bucket.create()

    response = bucket.generate_index_html("testing", "testing")
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200


@mock_s3
def test_get_bucket_cors_unset(bucket):
    bucket.create()

    with pytest.raises(NoSuchCORSConfiguration):
        bucket.get_cors_config()


@mock_s3
def test_set_bucket_cors(bucket):
    bucket.create()

    bucket.set_cors_config(origins=["testsite.com", "testsite2.ch"])
    response = bucket.get_cors_config()
    assert response["AllowedOrigins"] == ["testsite.com", "testsite2.ch"]


@mock_s3
def test_set_bucket_cors_allow_all(bucket):
    bucket.create()

    bucket.set_cors_config(allow_all=True)
    response = bucket.get_cors_config()
    assert response["AllowedOrigins"] == ["*"]


@mock_s3
def test_clean_multiparts(bucket, create_tempfile):
    bucket.create()

    client = Bucket.get_boto3_client()
    assert client

    key = "/test.txt"

    response = client.create_multipart_upload(
        Bucket=bucket.bucket_name,
        Key=key,
    )
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    with create_tempfile("txt") as upload:
        upload.write(b"0" * 1024 * 1024 * 5)  # 5MB

        response = client.upload_part(
            Bucket=bucket.bucket_name,
            Key=key,
            PartNumber=1,
            UploadId=response["UploadId"],
            Body=upload,
        )
        assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    response = client.list_multipart_uploads(
        Bucket=bucket.bucket_name,
    )
    assert len(response["Uploads"]) == 1

    status_dict = bucket.clean_multiparts()
    assert status_dict[key] is True

    response = client.list_multipart_uploads(
        Bucket=bucket.bucket_name,
    )
    assert "Uploads" not in response, "The multipart clean failed."


@mock_s3
def test_clean_multiparts_empty_bucket(bucket):
    bucket.create()

    status_dict = bucket.clean_multiparts()
    assert status_dict == {}


@mock_s3
def test_get_bucket_size(bucket, create_tempfile):
    bucket.create()

    with create_tempfile("txt") as temp1:
        with open(temp1.name, "w") as w:
            for _n in range(0, 57):
                w.write(str(1))
                w.write(",")

        status = bucket.upload_file(temp1.name, temp1.name)
        assert status is True

    bucket_size = bucket.size()
    assert bucket_size == 114


@mock_s3
def test_get_bucket_size_empty(bucket):
    bucket.create()

    bucket_size = bucket.size()
    assert bucket_size == 0


@mock_s3
def test_get_bucket_many_page(bucket, create_tempfile):
    bucket.create()

    with create_tempfile("txt") as temp1:
        with open(temp1.name, "w") as w:
            for _n in range(0, 20):
                w.write(str(1))
                w.write(",")

        status = bucket.upload_file(temp1.name, temp1.name)
        assert status is True

    bucket_size = bucket.size(items_per_page=5)
    assert bucket_size == 40


@mock_s3
def test_list_buckets(bucket, bucket2):
    bucket.create()
    bucket2.create()

    all_buckets = Bucket.list_buckets()
    assert all_buckets == ["testing", "testing2"]



"""DataCite for linking metadata to DOIs."""

import collections
import json
import os
import re
from datetime import date
from json import JSONDecodeError
from logging import getLogger

import jsonschema
import validators
from xmltodict import unparse

from envidat.utils import get_url, load_dotenv_if_in_debug_mode

log = getLogger(__name__)


# Load config from environment variables
load_dotenv_if_in_debug_mode(".env")

# TODO add return type hints to functions

def convert_datacite(metadata_record: dict) -> str | None:
    """Generate XML formatted string in DataCite format.

    Note:
        Converter is only valid for the metadata schema for EnviDat.

    Args:
        metadata_record (dict): EnviDat metadata entry record dictionary.

    Returns:
        str: XML formatted string compatible with DataCite DIF 10.2 standard
    """
    try:
        # Load and validate config
        config: dict = get_config_datacite_converter()
        if not config:
            return None

        # Convert metadata record to OrderedDict in DataCite format
        converted_package = datacite_convert_dataset(metadata_record, config)

        # Convert OrderedDict to XML
        if converted_package:
            return unparse(converted_package, pretty=True)
        else:
            log.error("ERROR failed to convert record to DataCite format, "
                      "check log for error causes")
            return None

    except TypeError as err:
        log.error(f"ERROR failed to convert record to DataCite format, "
                  f"error: {err}")
        return None


def get_config_datacite_converter(
        config_path: str = "envidat/config/config_converters.json"
) -> dict | None:
    """Return validated datacite converter JSON config as Python dictionary.

    Dictionary maps Datacite XML schema tags (keys) to EnviDat schema fields
    (values).

    Args:
        config_path (str): Path to JSON config file,
                        default path is "envidat/config/config_converters.json"

    Returns:
        dict: datacite converter JSON config as Python dictionary
        None: if config failed validation
    """
    with open(config_path, encoding="utf-8") as config_json:

        # Load config
        config: dict = json.load(config_json)
        datacite_config: dict = config["datacite_converter"]

        # Validate DataCite config has keys REQUIRED by DataCite Metadata
        # Schema 4.4,
        # for documentation see https://schema.datacite.org/meta/kernel-4.4/
        try:
            validate_dc_config(datacite_config)
            return datacite_config
        except jsonschema.exceptions.ValidationError as e:
            log.error(f"ERROR 'datacite_converter' config"
                      f" in '{config_path}' invalid:  {e}")
            return None


# TODO finish refactoring Datacite converter
# TODO keep in mind that a reverse converter will
#  also need to be written (DataCite to EnviDat)
# TODO refactor to pass env config to all helper functions instead
#  of loading config in helper functions
def datacite_convert_dataset(dataset: dict, config: dict):
    """Convert EnviDat metadata package from CKAN to DataCite XML.

    Notes: This converter is compatible with DataCite Metadata Schema 4.4, for
    documentation see https://schema.datacite.org/meta/kernel-4.4/

    Args:
    dataset (dict): EnviDat metadata entry record dictionary.
    config (dict): datacite converter config dervied from JSON config

    Returns:
    collections.orderedDict: ordered dictionary with input record converted to
    DataCite format
    """
    # Initialize ordered dictionary that will contain
    # dataset content converted to DataCite format
    dc = collections.OrderedDict()

    # Assign language tag used several times in function
    dc_xml_lang_tag = "xml:lang"

    # Header
    dc["resource"] = collections.OrderedDict()
    namespace = "http://datacite.org/schema/kernel-4"
    schema = "http://schema.datacite.org/meta/kernel-4.4/metadata.xsd"
    dc["resource"]["@xsi:schemaLocation"] = f"{namespace} {schema}"
    dc["resource"]["@xmlns"] = f"{namespace}"
    dc["resource"]["@xmlns:xsi"] = "http://www.w3.org/2001/XMLSchema-instance"

    # REQUIRED DataCite property: "Identifer",
    #   required attribute is "identifierType" (must be "DOI")
    dc_identifier_tag = "identifier"
    doi = dataset.get(config[dc_identifier_tag])
    if doi:
        dc["resource"][dc_identifier_tag] = {
            "#text": doi.strip(),
            "@identifierType": "DOI",
        }
    else:
        log_falsy_value(config[dc_identifier_tag])
        return None

    # REQUIRED DataCite property: "Creator"
    dc_creators_tag = "creators"
    dc_creator_tag = "creator"
    dc["resource"][dc_creators_tag] = {dc_creator_tag: []}
    author_dataset = dataset.get(config[dc_creators_tag], [])

    if author_dataset:
        try:
            authors = json.loads(author_dataset)
        except JSONDecodeError:
            log.error(
                f"ERROR cannot parse '{config[dc_creators_tag]}' value from package")
            return None
    else:
        log_falsy_value(config[dc_creators_tag])
        return None

    # REQUIRED DataCite attibute for each "creator" property: "creatorName"
    for author in authors:
        dc_creator = get_dc_creator(author, config)
        if dc_creator:
            if "creatorName" in dc_creator:
                dc["resource"][dc_creators_tag][dc_creator_tag] += [dc_creator]
            else:
                log.error(f"ERROR missing required 'name' value from "
                          f"'{config[dc_creators_tag]}' key in input record for author "
                          f"{author}")

    # Check if there is at least one Creator in dc
    creators = dc["resource"][dc_creators_tag][dc_creator_tag]
    if not creators:
        log.error(f"ERROR missing at least one required valid 'Creator' derived from "
                  f"'{config[dc_creators_tag]}' key in input record")
        return None

    # REQUIRED DataCite property: "title"
    dc_titles_tag = "titles"
    dc_title_tag = "title"
    dc["resource"][dc_titles_tag] = {dc_title_tag: []}
    title = dataset.get(config[dc_title_tag], "")
    if title:
        dc["resource"][dc_titles_tag][dc_title_tag] = {
            f"@{dc_xml_lang_tag}": "en-us",
            "#text": title,
        }
    else:
        log_falsy_value(config[dc_title_tag])
        return None

    # Get publication dictionary
    pub = dataset.get("publication", {})
    try:
        publication = json.loads(pub)
    except JSONDecodeError:
        publication = {}

    # REQUIRED DataCite property: "publisher" (default publisher is "EnviDat")
    dc_publisher_tag = "publisher"
    publisher = publication.get(config[dc_publisher_tag], "EnviDat")
    if not publisher:
        publisher = "EnviDat"

    if publisher:
        dc["resource"][dc_publisher_tag] = {
            f"@{dc_xml_lang_tag}": "en-us",
            "#text": publisher.strip(),
        }

    # REQUIRED DataCite property: "publicationYear" (default value is current year)
    dc_publication_year_tag = "publicationYear"
    publication_year = publication.get(config[dc_publication_year_tag], "")
    if publication_year:
        dc["resource"][dc_publication_year_tag] = {"#text": publication_year}
    else:
        publication_year = str(date.today().year)
        dc["resource"][dc_publication_year_tag] = {"#text": publication_year}

    # REQUIRED DataCite property: "resourceType" (default value is "dataset"),
    #   required attribute is "resourceTypeGeneral" (default value is "Dataset")
    dc_resource_type_tag = "resourceType"
    dc_resource_type_general_tag = "resourceTypeGeneral"
    resource_type_general = dataset.get(config[dc_resource_type_general_tag], "Dataset")
    dc_resource_type_general = value_to_datacite_cv(
        resource_type_general, dc_resource_type_general_tag, default="Dataset"
    )

    dc["resource"][dc_resource_type_tag] = {
        "#text": dataset.get(config[dc_resource_type_tag], "dataset"),
        f"@{dc_resource_type_general_tag}": dc_resource_type_general,
    }

    # Subjects
    dc_subjects_tag = "subjects"
    dc_subject_tag = "subject"
    dc_subjects = []

    tags = dataset.get(config[dc_subjects_tag], [])
    for tag in tags:
        tag_name = tag.get(config[dc_subject_tag], tag.get("name", ""))
        if tag_name:
            dc_subjects += [{f"@{dc_xml_lang_tag}": "en-us", "#text": tag_name}]

    if dc_subjects:
        dc["resource"][dc_subjects_tag] = {dc_subject_tag: dc_subjects}

    # Contributor (contact person)
    dc_contributors_tag = "contributors"
    dc_contributor_tag = "contributor"
    dc["resource"][dc_contributors_tag] = {dc_contributor_tag: []}

    # Get "maintainer" from EnviDat package,
    # assigned as DataCite Contributor "ContactPerson"
    maintainer_dataset = dataset.get(config[dc_contributors_tag], {})
    try:
        maintainer = json.loads(maintainer_dataset)
    except JSONDecodeError:
        maintainer = {}

    dc_contributor = get_dc_contributor(maintainer, config)
    if dc_contributor:
        dc["resource"][dc_contributors_tag][dc_contributor_tag] += [dc_contributor]

    # Get "organization" dataset and extract "name" value,
    # assigned as DataCite Contributor "ResearchGroup"
    organization = dataset.get("organization", {})
    if organization:
        organization_title = organization.get("title", "")
        if organization_title:
            dc_research_group = get_dc_research_group(organization_title)
            dc["resource"][dc_contributors_tag][dc_contributor_tag] \
                += [dc_research_group]

    # Dates
    dc_dates_tag = "dates"
    dc_date_tag = "date"
    dc_date_type_tag = "dateType"
    dc_dates = []

    date_input = dataset.get(config[dc_dates_tag], [])
    try:
        dates = json.loads(date_input)
    except JSONDecodeError:
        dates = []

    # "dateType" is REQUIRED DataCite attribute for each "Date", (default value is
    #    "Valid"), log values that are not "Created" or "colected"
    for dte in dates:

        date_type = (dte.get(config[dc_date_type_tag]))
        if date_type not in ["created", "Created", "collected", "Collected"]:
            log.warning(f"WARNING {config[dc_date_type_tag]} value '{date_type}' "
                        f"not a valid DataCite {dc_date_type_tag} ")
            date_type = "Valid"

        dc_date = {
            "#text": dte.get(config[dc_date_tag], ""),
            f"@{dc_date_type_tag}": date_type.title()
        }
        dc_dates += [dc_date]

    if dc_dates:
        dc["resource"][dc_dates_tag] = {dc_date_tag: dc_dates}

    # Language, "en" (English is default langauge)
    dc_language_tag = "language"
    dc_language = dataset.get(config[dc_language_tag], "")
    if not dc_language:
        dc_language = "en"
    dc["resource"][dc_language_tag] = {"#text": dc_language}

    # Alternate Identifier
    # "alternateIdentifierType" is a required attribute for each "alternateIdentifier",
    #   (value assigned to "URL")
    base_url = "https://www.envidat.ch/#/metadata/"
    alternate_identifiers = []

    package_name = dataset.get("name", "")
    if package_name:
        package_url = f"{base_url}{package_name}"
        alternate_identifiers.append(
            {"#text": package_url, "@alternateIdentifierType": "URL"}
        )

    package_id = dataset.get("id", "")
    if package_id:
        package_id = f"{base_url}{package_id}"
        alternate_identifiers.append(
            {"#text": package_id, "@alternateIdentifierType": "URL"}
        )

    dc["resource"]["alternateIdentifiers"] = {
        "alternateIdentifier": alternate_identifiers
    }

    # Related identifiers (from "related_datasets", "related_publications",
    # and "resources" values)

    # Get "related_datasets" from Envidat record
    related_datasets = dataset.get("related_datasets", "")
    dc_related_datasets = get_dc_related_identifiers(related_datasets,
                                                     has_related_datasets=True)

    # Get "related_publications" from EnviDat record
    related_publications = dataset.get("related_publications", "")
    dc_related_publications = get_dc_related_identifiers(related_publications)

    # Get "resources" from EnviDat record,
    # used for DataCite "relatedIdentifiers" and "formats" tags
    resources = dataset.get("resources", [])
    dc_resources = get_dc_related_identifiers_resources(resources)

    # Combine related identifiers from different sources
    related_ids = []
    related_id_sources = [
        dc_related_datasets,
        dc_related_publications,
        dc_resources]

    # Assign related_ids to sources that are truthy (not empty list)
    for source in related_id_sources:
        if source:
            related_ids += source

    # Assign related_identifier tag(s) to dc
    dc_related_identifiers = collections.OrderedDict()
    if related_ids:
        dc_related_identifiers["relatedIdentifier"] = related_ids
        dc["resource"]["relatedIdentifiers"] = dc_related_identifiers

    # Formats (from resources)
    dc_formats = get_dc_formats(resources)
    if dc_formats:
        dc_format_group_tag = "formats"
        dc_format_tag = "format"
        dc["resource"][dc_format_group_tag] = {dc_format_tag: dc_formats}

    # Version
    dc_version_tag = "version"
    dc_version = dataset.get(config[dc_version_tag], "")
    if dc_version:
        dc["resource"][dc_version_tag] = {"#text": dc_version}

    # Rights
    dc_rights_group_tag = "rightsList"
    dc_rights_tag = "rights"
    dc_rights_uri_tag = "rightsURI"

    dc_scheme_uri_tag = "schemeURI"
    default_rights_scheme_uri = "https://spdx.org/licenses/"

    dc_rights_identifier_scheme = "rightsIdentifierScheme"
    default_rights_identifier = "SPDX"

    dc_rights_identifier = "rightsIdentifier"  # "CC0 1.0"

    rights = {}

    dc_rights_text = "#text"
    rights_title = dataset.get(config[dc_rights_tag][dc_rights_text], "")
    if rights_title:
        rights = {f"@{dc_xml_lang_tag}": "en-us", "#text": rights_title}

    rights_uri = dataset.get(config[dc_rights_tag][dc_rights_uri_tag], "")
    if rights_uri:
        rights[f"@{dc_rights_uri_tag}"] = rights_uri

    license_id = dataset.get(config[dc_rights_tag][dc_rights_identifier], "")
    # TODO investigate default argument
    rights_id_spx = value_to_datacite_cv(license_id, dc_rights_identifier, default=None)
    if rights_id_spx:
        rights[f"@{dc_scheme_uri_tag}"] = default_rights_scheme_uri
        rights[f"@{dc_rights_identifier_scheme}"] = default_rights_identifier
        rights[f"@{dc_rights_identifier}"] = rights_id_spx

    if rights:
        dc["resource"][dc_rights_group_tag] = {dc_rights_tag: [rights]}

    # Description
    dc_descriptions_tag = "descriptions"
    dc_description_tag = "description"
    dc_description_type_tag = "descriptionType"

    notes = dataset.get(config[dc_description_tag], "")
    dc_descriptions = get_dc_descriptions(
        notes, dc_description_type_tag, dc_xml_lang_tag
    )

    if dc_descriptions:
        dc["resource"][dc_descriptions_tag] = {dc_description_tag: dc_descriptions}
    else:
        log.warning(
            f"WARNING dataset does not have a truthy value for metadata record key: "
            f"{config[dc_description_tag]} ")

    # GeoLocation
    dc_geolocations_tag = "geoLocations"
    dc_geolocations = []

    # Get spatial data from dataset
    try:
        spatial = json.loads(dataset.get(config[dc_geolocations_tag], ""))
        spatial_type = spatial.get("type", "").lower()

        if spatial and spatial_type:
            if spatial_type == "geometrycollection":
                dc_geolocations = geometrycollection_to_dc_geolocations(spatial)
            else:
                dc_geolocations = get_dc_geolocations(spatial, spatial_type)
    except JSONDecodeError:
        dc_geolocations = []

    # Assign converted spatial and spatial_info values to corresponding DataCite tags
    if dc_geolocations:
        dc_geolocation_place_tag = "geoLocationPlace"

        geolocation_place = dataset.get(config[dc_geolocation_place_tag], "")
        if geolocation_place:
            datacite_geolocation_place = {
                dc_geolocation_place_tag: geolocation_place.strip()
            }
            dc_geolocations += [datacite_geolocation_place]

        dc["resource"][dc_geolocations_tag] = {"geoLocation": dc_geolocations}

    # Funding Information
    dc_funding_refs_tag = "fundingReferences"
    dc_funding_ref_tag = "fundingReference"

    funding_dataset = dataset.get(config[dc_funding_refs_tag], [])
    try:
        funding = json.loads(funding_dataset)
    except JSONDecodeError:
        funding = []

    dc_funding_refs = []

    for funder in funding:

        dc_funding_ref = collections.OrderedDict()
        dc_funder_name_tag = "funderName"

        # "funderName" is a REQUIRED DataCite attribute for each "fundingReference"
        funder_name = funder.get(config[dc_funding_ref_tag][dc_funder_name_tag], "")
        if funder_name:
            dc_funding_ref[dc_funder_name_tag] = funder_name.strip()

            dc_award_number_tag = "awardNumber"
            award_number = funder.get(
                config[dc_funding_ref_tag][dc_award_number_tag], ""
            )

            dc_award_uri_tag = "awardURI"
            award_uri = funder.get(config[dc_funding_ref_tag][dc_award_uri_tag], "")

            # Assign awardNumber and awardURI if they exist
            # and if awardURI is a valid URL
            # NOTE: For reverse converter be sure to parse default value for
            # awardNumber, ":unav"
            # DataCite documentation for unknown information: p. 74
            # https://schema.datacite.org/meta/kernel-4.4/doc/DataCite
            # -MetadataKernel_v4.4.pdf
            if award_uri and validators.url(award_uri):
                if award_number:
                    award = {f"@{dc_award_uri_tag}": award_uri,
                             "#text": award_number.strip()}
                else:
                    award = {f"@{dc_award_uri_tag}": award_uri, "#text": ":unav"}
                dc_funding_ref[dc_award_number_tag] = award
            elif award_number:
                dc_funding_ref[dc_award_number_tag] = award_number.strip()

            dc_funding_refs += [dc_funding_ref]

    if dc_funding_refs:
        dc["resource"][dc_funding_refs_tag] = {
            dc_funding_ref_tag: dc_funding_refs
        }

    return dc


def get_dc_creator(author: dict, config: dict):
    """Returns author information in DataCite "creator" tag format."""
    dc_creator_tag = "creator"
    dc_creator = collections.OrderedDict()

    creator_family_name = author.get(config[dc_creator_tag]["familyName"], "").strip()
    creator_given_name = author.get(config[dc_creator_tag]["givenName"], "").strip()

    if creator_given_name and creator_family_name:
        dc_creator["creatorName"] = f"{creator_given_name} {creator_family_name}"
        dc_creator["givenName"] = creator_given_name
        dc_creator["familyName"] = creator_family_name
    elif creator_family_name:
        dc_creator["creatorName"] = creator_family_name

    # REQUIRED DataCite property for each "Creator"
    # with a "nameIdentifier": "nameIdentifierScheme" (value assigned to "ORCID")
    creator_identifier = author.get(config[dc_creator_tag]["nameIdentifier"], "")
    if creator_identifier:
        dc_creator["nameIdentifier"] = {
            "#text": creator_identifier.strip(),
            "@nameIdentifierScheme": "ORCID",
            "@schemeURI": "https://orcid.org/"
        }

    affiliations = []
    affiliation = author.get(config[dc_creator_tag]["affiliation"], "")
    if affiliation:
        aff = affiliation_to_dc(affiliation, config)
        if aff:
            affiliations += [aff]

    affiliation_02 = author.get("affiliation_02", "")
    if affiliation_02:
        aff_02 = affiliation_to_dc(affiliation_02, config)
        if aff_02:
            affiliations += [aff_02]

    affiliation_03 = author.get("affiliation_03", "")
    if affiliation_03:
        aff_03 = affiliation_to_dc(affiliation_03, config)
        if aff_03:
            affiliations += [aff_03]

    if affiliations:
        dc_creator["affiliation"] = affiliations

    return dc_creator


def get_dc_contributor(maintainer: dict, config: dict):
    """Returns maintainer in DataCite "contributor" tag format with a
    contributorType of "ContactPerson".

    REQUIRED DataCite attribute for each "contributor": "contributorType",
                                      (value assigned is "Contact Person")

    REQUIRED DataCite property for each "contibutor": "contributorName"

    REQUIRED DataCite property for each "nameIdentifier" property:
                                       "nameIdentifierScheme" (default value is "ORCID")
    """
    dc_contributor = collections.OrderedDict()
    dc_contributor_tag = "contributor"

    contributor_family_name = maintainer.get(
        config[dc_contributor_tag]["familyName"], ""
    ).strip()
    contributor_given_name = maintainer.get(
        config[dc_contributor_tag]["givenName"], ""
    ).strip()

    if contributor_given_name:
        dc_contributor[
            "contributorName"] = f"{contributor_given_name} {contributor_family_name}"
        dc_contributor["givenName"] = contributor_given_name
        dc_contributor["familyName"] = contributor_family_name
    else:
        dc_contributor["contributorName"] = contributor_family_name

    contributor_identifier = maintainer.get(
        config[dc_contributor_tag]["nameIdentifier"], ""
    )
    if contributor_identifier:
        dc_contributor["nameIdentifier"] = {
            "#text": contributor_identifier.strip(),
            "@nameIdentifierScheme": maintainer.get(
                join_tags(
                    [dc_contributor_tag, "nameIdentifier", "nameIdentifierScheme"]
                ),
                "orcid",
            ).upper(),
            "@schemeURI": "https://orcid.org/"
        }

    contributor_affiliation = maintainer.get(
        config[dc_contributor_tag]["affiliation"], ""
    )

    if contributor_affiliation:
        affiliation_dc = affiliation_to_dc(contributor_affiliation, config)
        if affiliation_dc:
            dc_contributor["affiliation"] = affiliation_dc

    contributor_type = maintainer.get(
        join_tags([dc_contributor_tag, "contributorType"]), "ContactPerson"
    )
    dc_contributor["@contributorType"] = value_to_datacite_cv(
        contributor_type, "contributorType"
    )

    return dc_contributor


def affiliation_to_dc(affiliation, config) -> dict[str, str]:
    """Returns affiliation in DataCite "affiliation" tag format.

    Uses config to map commonly used affiliations in EnviDat packages
    (i.e. "WSL", "SLF") with long names of instiutions
    and ROR identifiers when available.
    """
    # Get key from config that corresponds to affiliation
    aff_keys = {
        "WSL": "wsl",
        "Swiss Federal Institute for Forest, Snow and Landscape Research WSL": "wsl",
        "WSL Swiss Federal Research Institute, Birmensdorf, Switzerland": "wsl",
        "SLF": "slf",
        "WSL Institute for Snow and Avalanche Research SLF, Davos Dorf, Switzerland":
            "slf",
        "WSL Institute for Snow and Avalanche Research SLF": "slf",
        "ETH": "eth",
        "ETHZ": "eth",
        "UZH": "uzh",
        "University of Zurich": "uzh",
        "University of Zürich": "uzh",
        "EPFL": "epfl",
        "EPFL, Lausanne Swiss Federal Institute of Technology, Lausanne and Sion":
            "epfl",
        "PSI": "psi",
        "PSI, Paul Scherrer Institute, Villigen": "psi",
        "IAP": "iap",
        "TROPOS": "tropos",
        "UNIL": "unil"
    }

    # Get affiliation config
    aff_config = config["affiliation"]

    # Strip whitespace from affiliation
    aff = affiliation.strip()

    # Return org dictionary if it exists in config
    aff_key = aff_keys.get(aff, "")
    org = aff_config.get(aff_key, {})
    if org:
        # If "affiliationIdentifier" exists then "affiliationIdentifierScheme" REQUIRED
        # DataCite attibute
        if "@affiliationIdentifier" in org and "@affiliationIdentifierScheme" not in \
                org:
            log.warning(
                f"WARNING missing required '@affiliationIdentifierScheme' "
                f"key from config for affiliation: '{aff_key}'")
            return {"#text": aff}
        else:
            return org

    # Else return only affiliation
    return {"#text": aff}


def get_dc_research_group(organization_title):
    """Returns organization title in DataCite "contributor" format with a
    contributorType of "ResearchGroup".
    """
    dc_contributor = collections.OrderedDict()

    dc_contributor["@contributorType"] = "ResearchGroup"

    dc_contributor["contributorName"] = {
        "#text": organization_title.strip(),
        "@nameType": "Organizational"
    }

    return dc_contributor


def get_dc_related_identifiers(related_identifiers: str,
                               has_related_datasets=False) -> list[dict[str, str]]:
    """Return EnviDat records "related_datasets" or "related_publications" values in
    DataCite "relatedIdentifiers" tag format.

    Note:
        "relatedIdentiferType" and "relationType" are required attributes
        for each "relatedIdentifer"

    Args:
        related_identifiers (str): Input related idetifiers, expected input is from
            "related_datasets" or "related_publications" keys.
        has_related_datasets (bool): If true then input is assumed to be from
            "related_datasets" value in EnviDat record.
            Default value is false and is assumed to correspond to
            "related_publications" value in EnviDat record.
    """
    # Assign relation_type
    if has_related_datasets:
        # Corresponds to "related_datasets" key
        relation_type = "Cites"
    else:
        # Corresponds to "related_publications" key
        relation_type = "IsSupplementTo"

    # Assign empty list to contain related identifiers
    dc_related_identifiers = []

    # Validate related_identifiers
    if related_identifiers:

        # Remove special characters "\r", "\n" and
        # remove Markdown link syntax using brackets and parentheses
        # and replace with one space " "
        related_identifiers = re.sub(r"\r|\n|\[|\]|\(|\)", " ", related_identifiers)

        # Assign empty array to hold "related_ids" values that will be used to check for
        # duplicates
        related_ids = []

        # Extract DOIs
        for word in related_identifiers.split(" "):

            # Apply search function to find DOIs
            doi = get_doi(word)

            # If not doi then apply DORA API DOI search function
            if not doi:
                doi = get_dora_doi(word)

            # If not doi then apply EnviDat CKAN API DOI search function
            if not doi:
                doi = get_envidat_doi(word)

            # Add doi to dc_related_identifiers if it meets conditions
            if doi and "/" in doi and doi not in related_ids:
                related_ids.append(doi)

                dc_related_identifiers += [
                    {
                        "#text": doi,
                        "@relatedIdentifierType": "DOI",
                        "@relationType": relation_type,
                    }
                ]

                continue

            # Apply URL validator to find other URLs (that are not DORA or EnviDat DOIs)
            is_url = validators.url(word)

            # Add URL to dc_related_identifiers if it meets conditions
            if all([is_url,
                    word not in related_ids,
                    "doi" not in word,
                    "dora.lib4ri.ch/wsl/islandora/object/" not in word]):
                related_ids.append(word)

                dc_related_identifiers += [
                    {
                        "#text": word,
                        "@relatedIdentifierType": "URL",
                        "@relationType": relation_type,
                    }
                ]

    return dc_related_identifiers


def get_dc_related_identifiers_resources(resources) -> list[dict[str, str]]:
    """Return URLs from resources in DataCite "relatedIdentifier" tag format.

    Note:
        "relatedIdentiferType" and "relationType" are required attributes
        for each "relatedIdentifer"
    """
    dc_related_identifier = []

    for resource in resources:
        resource_url = resource.get("url", "")
        if resource_url:
            dc_related_identifier += [
                {
                    "#text": resource_url,
                    "@relatedIdentifierType": "URL",
                    "@relationType": "IsRequiredBy",
                }
            ]

    return dc_related_identifier


def get_dc_formats(resources) -> list[dict[str, str]]:
    """Returns resources formats in DataCite "formats" tag format."""
    dc_formats = []

    for resource in resources:

        default_format = resource.get("mimetype", resource.get("mimetype_inner", ""))
        resource_format = resource.get("format", "")

        if not resource_format:
            resource_format = default_format

        if resource_format:
            dc_format = {"#text": resource_format}
            dc_formats += [dc_format]

    return dc_formats


def get_dc_descriptions(notes, dc_description_type_tag, dc_xml_lang_tag) -> list[str]:
    """Returns notes in DataCite "descriptions" tag format.

    "descriptionType" is a REQUIRED DataCite attribute for each "description",
         (value assigned to "Abstract")

    Logs warning for a description that is less than 100 characters.
    """
    dc_descriptions = []

    if notes:
        description_text = (
            notes.replace("\r", "")
            .replace(">", "-")
            .replace("<", "-")
            .replace("__", "")
            .replace("#", "")
            .replace("\n\n", "\n")
            .replace("\n\n", "\n")
        )

        datacite_description = {
            "#text": description_text.strip(),
            f"@{dc_description_type_tag}": "Abstract",
            f"@{dc_xml_lang_tag}": "en-us",
        }

        if len(description_text) < 100:
            log.warning(
                f"WARNING description is less than 100 characters: {description_text}")

        dc_descriptions += [datacite_description]

    return dc_descriptions


def geometrycollection_to_dc_geolocations(spatial: dict):
    """Returns spatial data in DataCite "geoLocations" format.

    Assumption: input spatial dictionary has a "type" value of "geometrycollection".
    """
    dc_geolocations = []

    geometries = spatial.get("geometries")
    if geometries:

        for geometry in geometries:
            spatial_type = geometry.get("type", "")

            if spatial_type:
                dc_geolocation = get_dc_geolocations(geometry, spatial_type)

                if dc_geolocation:
                    dc_geolocations += dc_geolocation

    return dc_geolocations


# TODO test refactored get_dc_geolocations()
def get_dc_geolocations(spatial: dict, spatial_type: str = ""):
    """Returns spatial data in DataCite "geoLocations" format.

    For list of required attributes for each type of GeoLocation see DataCite documentation.
    """
    dc_geolocations = []

    spatial_type = spatial_type.lower()
    coordinates = spatial.get("coordinates")

    if coordinates and spatial_type:

        match spatial_type:

            case "polygon":
                dc_geolocation = get_dc_geolocation_polygon(coordinates)
                if dc_geolocation:
                    dc_geolocations += [dc_geolocation]

            case "point":
                dc_geolocation = get_dc_geolocation_point(coordinates)
                if dc_geolocation:
                    dc_geolocations += [dc_geolocation]

            case "multipoint":
                for coordinates_pair in coordinates:
                    dc_geolocation = get_dc_geolocation_point(coordinates_pair)
                    if dc_geolocation:
                        dc_geolocations += [dc_geolocation]

    return dc_geolocations


def get_dc_geolocation_polygon(coordinates: list):
    """Returns spatial data in DataCite "geoLocationPolygon" format.

    Returns None if coordinates invalid or < 4 coordinates_pairs obtained

    Limitation: Only can process first list in coordinates list from parsed geojson.
                This means that polygons with "holes" are not supported.
    """
    # Log warning if coordinates has more than one element (i.e. polygon with "hole")
    if len(coordinates) > 1:
        log.warning(f"Input 'spatial' data polygon has more than one list in coodinates from parsed geojson. "
                    f"Polygons with 'holes' are not currently supported. Coordinates:  {coordinates}")

    # Assign polygon_coordinates to first element of coordinates list
    polygon_coordinates = coordinates[0]

    # Validate polygon_coordinates
    if len(polygon_coordinates) < 4:
        log.warning(f"Input 'spatial' data polygon does not have at least 4 coordinates. Coordinates:  {coordinates}")
        return None
    if polygon_coordinates[0] != polygon_coordinates[-1]:
        log.warning(f"Input 'spatial' data polygon's is not a valid polygon becase the first "
                    f"point is not identical to the last point. Coordinates:  {coordinates}")
        return None

    # Convert input coordinates to DataCite format
    dc_geolocation = collections.OrderedDict()
    dc_gelocation_polygon_tag = "geoLocationPolygon"
    dc_geolocation[dc_gelocation_polygon_tag] = {"polygonPoint": []}

    for coordinates_pair in polygon_coordinates:
        if len(coordinates_pair) == 2:
            geolocation_point = collections.OrderedDict()
            geolocation_point["pointLongitude"] = coordinates_pair[0]
            geolocation_point["pointLatitude"] = coordinates_pair[1]
            dc_geolocation[dc_gelocation_polygon_tag]["polygonPoint"] += [geolocation_point]

    if dc_geolocation:
        return dc_geolocation

    return None


def get_dc_geolocation_point(coordinates_pair: list[float]):
    """Returns spatial data in DataCite's "geoLocationPoint" format.

    If coordinates_pair list does not have a length of two then returns None.
    """
    if len(coordinates_pair) == 2:
        dc_geolocation = collections.OrderedDict()
        dc_geolocation_point_tag = "geoLocationPoint"
        dc_geolocation[dc_geolocation_point_tag] = collections.OrderedDict()

        dc_geolocation[dc_geolocation_point_tag]["pointLongitude"] = coordinates_pair[0]
        dc_geolocation[dc_geolocation_point_tag]["pointLatitude"] = coordinates_pair[1]

        return dc_geolocation

    return None


# TODO remove legacy get_dc_geolocations() function after testing replacement
# def get_dc_geolocations(spatial: dict):
#     """Returns spatial data in DataCite "geoLocations" format
#
#     For list of required attributes for each type of GeoLocation
#     see DataCite documentation.
#     """
#
#     dc_geolocations = []
#
#     if spatial.get("type", "").lower() == "polygon":
#         dc_geolocation = collections.OrderedDict()
#         dc_gelocation_polygon_tag = "geoLocationPolygon"
#         dc_geolocation[dc_gelocation_polygon_tag] = {"polygonPoint": []}
#
#         for coordinates_pair in spatial.get("coordinates", "[[]]")[0]:
#             geolocation_point = collections.OrderedDict()
#             geolocation_point["pointLongitude"] = coordinates_pair[0]
#             geolocation_point["pointLatitude"] = coordinates_pair[1]
#             dc_geolocation[dc_gelocation_polygon_tag]["polygonPoint"] += [
#                 geolocation_point]
#
#         dc_geolocations += [dc_geolocation]
#
#     else:
#         dc_geolocation_point_tag = "geoLocationPoint"
#
#         if spatial.get("type", "").lower() == "multipoint":
#
#             for coordinates_pair in spatial.get("coordinates", "[]"):
#                 dc_geolocation = collections.OrderedDict()
#                 dc_geolocation[dc_geolocation_point_tag] = collections.OrderedDict()
#                 dc_geolocation[dc_geolocation_point_tag]["pointLongitude"] = \
#                     coordinates_pair[0]
#                 dc_geolocation[dc_geolocation_point_tag]["pointLatitude"] = \
#                     coordinates_pair[1]
#                 dc_geolocations += [dc_geolocation]
#
#         elif spatial.get("type", "").lower() == "geometrycollection":
#             pass
#         # elif spatial.get("type", "").lower() == "geometrycollection":
#         #
#         #     for geometry in spatial.get("geometries"):
#         #
#         #         if geometry.get("type", "").lower() == "point":
#         #             coordinates = geometry.get("coordinates")
#         #
#         #             dc_geolocation = collections.OrderedDict()
#         #             dc_geolocation[dc_geolocation_point_tag] =
#         #             collections.OrderedDict()
#         #
#         #             dc_geolocation[dc_geolocation_point_tag]["pointLongitude"] = \
#         #                 coordinates[0]
#         #             dc_geolocation[dc_geolocation_point_tag]["pointLatitude"] = \
#         #                 coordinates[1]
#         #             dc_geolocations += [dc_geolocation]
#
#         else:
#             dc_geolocation = collections.OrderedDict()
#             dc_geolocation[dc_geolocation_point_tag] = collections.OrderedDict()
#
#             coordinates = flatten(spatial.get("coordinates", "[]"), reverse=True)
#             dc_geolocation[dc_geolocation_point_tag]["pointLongitude"] = coordinates[1]
#             dc_geolocation[dc_geolocation_point_tag]["pointLatitude"] = coordinates[0]
#             dc_geolocations += [dc_geolocation]
#
#     return dc_geolocations


def flatten(inp: list, reverse: bool = False) -> list:
    """Flatten list, i.e. remove a dimension/nesting."""
    output = []
    for item in inp:
        if type(item) is not list:
            if reverse:
                output = [str(item)] + output
            else:
                output += [str(item)]
        else:
            output += flatten(item, reverse)
    return output


def join_tags(tags: list, sep: str = ".") -> str:
    """Join tags by a provided separator."""
    return sep.join([tag for tag in tags if tag])


def value_to_datacite_cv(value: str, datacite_tag: str, default: str = "") -> dict:
    """Constant definitions."""
    datacite_cv = {
        "titleType": {
            "alternativetitle": "AlternativeTitle",
            "subtitle": "Subtitle",
            "translatedtitle": "TranslatedTitle",
            "other": "Other",
        },
        "resourceTypeGeneral": {
            "audiovisual": "Audiovisual",
            "collection": "Collection",
            "dataset": "Dataset",
            "event": "Event",
            "image": "Image",
            "interactiveresource": "InteractiveResource",
            "model": "Model",
            "physicalobject": "PhysicalObject",
            "service": "Service",
            "software": "Software",
            "sound": "Sound",
            "text": "Text",
            "workflow": "Workflow",
            "other": "Other",
        },
        "descriptionType": {
            "abstract": "Abstract",
            "methods": "Methods",
            "seriesinformation": "SeriesInformation",
            "tableofcontents": "TableOfContents",
            "other": "Other",
        },
        "contributorType": {
            "contactperson": "ContactPerson",
            "datacollector": "DataCollector",
            "datacurator": "DataCurator",
            "datamanager": "DataManager",
            "distributor": "Distributor",
            "editor": "Editor",
            "funder": "Funder",
            "hostinginstitution": "HostingInstitution",
            "other": "Other",
            "producer": "Producer",
            "projectleader": "ProjectLeader",
            "projectmanager": "ProjectManager",
            "projectmember": "ProjectMember",
            "registrationagency": "RegistrationAgency",
            "registrationauthority": "RegistrationAuthority",
            "relatedperson": "RelatedPerson",
            "researchgroup": "ResearchGroup",
            "rightsholder": "RightsHolder",
            "researcher": "Researcher",
            "sponsor": "Sponsor",
            "supervisor": "Supervisor",
            "workpackageleader": "WorkPackageLeader",
        },
        "rightsIdentifier": {
            "odc-odbl": "ODbL-1.0",
            "cc-by-sa": "CC-BY-SA-4.0",
            "cc-by-nc": "CC-BY-NC-4.0",
        },
    }

    # Matching ignoring blanks, case, symbols
    value_to_match = value.lower().replace(" ", "").replace("_", "")
    match_cv = datacite_cv.get(datacite_tag, {}).get(value_to_match, default)

    return match_cv


def get_doi(word: str) -> str | None:
    """Get DOI string from input word string, if DOI not found then returns None.

    Example:
        an input of "https://doi.org/10.1525/cse.2022.1561651" would return
        "10.1525/cse.2022.1561651" as output

    Args:
        word (str): Input string to test if it contains a DOI

    Returns:
        str: String of DOI
        None: If DOI could not be found
    """
    doi = None

    # Apply search criteria to find DOIs
    search_strings = ["doi", "10."]
    if any(search_str in word for search_str in search_strings):

        # Assign doi if "10." in word
        doi_start_index = word.find("10.")
        if doi_start_index != -1:
            doi = word[doi_start_index:]

            # Remove unwanted trailing period characters if they exist
            unwanted_chars = [".", ","]
            if any(doi[-1] == char for char in unwanted_chars):
                doi = doi[:-1]

    # Return DOI if it exists, else return None
    return doi


# TODO possibly pass env config as argument
def get_envidat_doi(word: str,
                    api_host="https://envidat.ch",
                    api_package_show="/api/action/package_show?id=") -> str | None:
    """Get DOI string from input work by calling EnviDat API,
         if DOI not found then returns None.

    Example:
        An input of
        "https://www.envidat.ch/#/metadata/amphibian-and-landscape-data-swiss-lowlands"
        would return ""10.16904/envidat.267" as output

    Args:
        word (str): Input string to test if it contains a DOI retrieved from
                    EnviDat CKAN API
        api_host (str): API host URL. Attempts to get from environment.
            Default value is "https://envidat.ch".
        api_package_show (str): API host path to show package. Attempts to get from
             environment. Default value is "/api/action/package_show?id="

    Returns:
        str: String of DOI
        None: If DOI could not be found
    """
    doi = None

    # Check if word meets search criteria to be an EnviDat package URL
    if word.startswith(
            ("https://www.envidat.ch/#/metadata/", "https://www.envidat.ch/dataset/")) \
            and "/resource/" not in word:

        # Extract package_name from package URL
        last_slash_index = word.rfind("/")
        if last_slash_index != -1:
            package_name = word[(last_slash_index + 1):]

            # Extract environment variables from config, else use default values
            api_host = os.getenv("API_HOST", default=api_host)
            api_package_show = os.getenv("API_PACKAGE_SHOW", default=api_package_show)

            # Assemble URL used to call EnviDat CKAN API
            api_url = f"{api_host}{api_package_show}{package_name}"

            # Call API and try to return doi
            try:
                data = get_url(api_url).json()
                if data:
                    doi = data.get("result").get("doi")
                    if doi:
                        return doi
            except Exception as e:
                log.error(f"ERROR: Failed to retrieve'{api_url}' and extract DOI")
                log.error(e)
                return None

    return doi


def get_dora_doi(word: str) -> str | None:
    """Get DOI string from input word string by calling DORA API,
         if DOI not found then returns None.

    Example:
        an input of "https://www.dora.lib4ri.ch/wsl/islandora/object/wsl%3A3213"
        would return "10.5194/tc-10-1075-2016" as output.

    Args:
        word (str): Input string to test if it contains a DOI retrieved from DORA API

    Returns:
        str: String of DOI
        None: If DOI could not be found
    """
    doi = None

    # Apply search criteria to find DOIs from DORA API
    # DORA API documentation:
    # https://www.wiki.lib4ri.ch/display/HEL/Technical+details+of+DORA
    dora_str = "dora.lib4ri.ch/wsl/islandora/object/"
    if dora_str in word:
        dora_start_index = word.find(dora_str)
        dora_pid = word[(dora_start_index + len(dora_str)):]

        # Remove any characters that may exist after DORA PID
        dora_end_index = dora_pid.find("/")

        # Modify dora_pid if dora_end_index found in dora_pid
        if dora_end_index != -1:
            dora_pid = dora_pid[:dora_end_index]

        # Call DORA API and get DOI if it listed in citation
        doi_dora = get_dora_doi_string(dora_pid)
        if doi_dora:
            doi = doi_dora

    return doi


# TODO possibly pass env config as argument
def get_dora_doi_string(
        dora_pid: str, dora_api_url: str = "https://envidat.ch/dora") -> str | None:
    """Get DOI string from WSL DORA API using DORA PID.

    DORA API documentation:
    https://www.wiki.lib4ri.ch/display/HEL/Technical+details+of+DORA

    ASSUMPTION: Only one DOI exists in each DORA API record "citation" key

    Args:
        dora_pid (str): DORA PID (permanent identification)
        dora_api_url (str): API host url. Attempts to get from environment.
            Defaults to "https://envidat.ch/dora"

    Returns:
        str: String of DOI
        None: If DOI could not be found
    """
    # Extract environment variables from config, else use default values
    dora_api_url = os.getenv("DORA_API_URL", default=dora_api_url)

    # Replace '%3A' ASCII II code with semicolon ':'
    dora_pid = re.sub("%3A", ":", dora_pid)

    # Replace literal asterisk "\*" with empty string ""
    dora_pid = re.sub("\\*", "", dora_pid)

    # Assemble url used to call DORA API
    dora_url = f"{dora_api_url}/{dora_pid}"

    try:
        data = get_url(dora_url).json()

        if data:
            citation = data[dora_pid]["citation"]["WSL"]

            for word in citation.split(" "):
                # Return DOI if it exists
                doi = get_doi(word)
                if doi:
                    return doi

            # If DOI not found then return None
            return None

        return None

    except Exception as e:
        log.error(f"ERROR: Failed to retrieve'{dora_url}' and extract DOI")
        log.error(e)
        return None


def log_falsy_value(key: str):
    """Logs error message for a falsy value from a EnviDat key that corresponds
    to a required DataCite property.

     Notes: This function used in datacite_convert_dataset() for required properties
     according toDataCite Metadata Schema 4.4, for documentation see
      https://schema.datacite.org/meta/kernel-4.4/

    Args:
        key (str): key that does not have truthy value from input EnviDat record
    """
    log.error(
        f"ERROR input record does not have truthy value for key '{key}', "
        f"this key corresponds to a required DataCite property")


def validate_dc_config(datacite_config: dict):
    """Validate DataCite config has DataCite required keys using jsonschema.

    Note:
        There are other DataCite required properties not included in this schema that
        are handled differently in the converter (such as using default values).

    Returns jsonschema.exceptions.ValidationError if input config invalid with schema.

    Args:
        datacite_config (dict): dictionary derived from "datacite_converter"
                                object in JSON config

    Returns:
        None: if datacite_config is valid against schema
        jsonschema.exceptions.ValidationError: if datacite_config is invalid
    """
    dc_schema = {
        "type": "object",
        "properties": {
            "identifier": {"type": "string"},
            "creators": {"type": "string"},
            "title": {"type": "string"}
        },
        "required": ["identifier", "creators", "title"],
    }

    jsonschema.validate(instance=datacite_config, schema=dc_schema)



"""Configuration file for PyTest tests."""

import os
from tempfile import NamedTemporaryFile
from textwrap import dedent

import pytest
from moto import mock_s3

from envidat.converters.bibtex_converter import bibtex_convert_dataset
from envidat.converters.datacite_converter import convert_datacite
from envidat.converters.dcat_ap_converter import dcat_ap_convert_dataset
from envidat.converters.dif_converter import dif_convert_dataset
from envidat.converters.iso_converter import iso_convert_dataset
from envidat.converters.ris_converter import ris_convert_dataset
from envidat.s3.bucket import Bucket

os.environ["MOTO_ALLOW_NONEXISTENT_REGION"] = "True"


# @pytest.fixture(scope="session")
# def s3_env_vars():

#     # Disable region validation for moto
#     os.environ["MOTO_ALLOW_NONEXISTENT_REGION"] = "True"

#     # Official vars
#     os.environ["AWS_ACCESS_KEY_ID"] = "testing"
#     os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
#     os.environ["AWS_SECURITY_TOKEN"] = "testing"
#     os.environ["AWS_SESSION_TOKEN"] = "testing"
#     os.environ["AWS_DEFAULT_REGION"] = "testing"

#     # Custom vars
#     os.environ["AWS_ENDPOINT"] = "testing"
#     os.environ["AWS_ACCESS_KEY"] = "testing"
#     os.environ["AWS_SECRET_KEY"] = "testing"
#     os.environ["AWS_REGION"] = "testing"
#     os.environ["AWS_BUCKET_NAME"] = "testing"


@pytest.fixture(scope="session")
@mock_s3
def bucket():
    """Bucket for tests."""
    Bucket.config("testing", "testing", endpoint=None, region="testing")
    new_bucket = Bucket("testing")
    return new_bucket


@pytest.fixture(scope="session")
@mock_s3
def bucket2():
    """Second bucket when two are required in tests."""
    Bucket.config("testing", "testing", endpoint=None, region="testing")
    new_bucket = Bucket("testing2")
    return new_bucket


@pytest.fixture
def create_tempfile(scope="function"):
    """Create temporary file in tests."""

    def nested_tempfile(file_type, temp_dir=None, delete=True):
        """Nested temporary file within subdirectory."""
        temp_file = NamedTemporaryFile(
            dir=temp_dir, delete=delete, suffix=f".{file_type}"
        )
        with open(temp_file.name, "w", encoding="UTF-8") as f:
            f.write("test")
        return temp_file

    return nested_tempfile


@pytest.fixture
def bibtex_converter_one_package():
    """Single package in BibTeX format."""
    package_name = "bioclim_plus"
    file_format = "bibtex"
    extension = "bib"
    return bibtex_convert_dataset, package_name, file_format, extension


@pytest.fixture
def bibtex_converter_all_packages():
    """All packages in BibTeX format."""
    file_format = "bibtex"
    extension = "bib"
    return bibtex_convert_dataset, file_format, extension


#
@pytest.fixture
def datacite_converter_one_package():
    """Single package in Datacite format."""
    package_name = "10-16904-3"
    file_format = "datacite"
    extension = "xml"
    return (
        convert_datacite,
        package_name,
        file_format,
        extension,
    )


@pytest.fixture
def datacite_converter_all_packages():
    """All packages in Datacite format."""
    file_format = "datacite"
    extension = "xml"
    return convert_datacite, file_format, extension


@pytest.fixture
def dif_converter_one_package():
    """Single package in Diff format."""
    package_name = "resolution-in-sdms-shapes-plant-multifaceted-diversity"
    file_format = "gcmd_dif"
    extension = "xml"
    return dif_convert_dataset, package_name, file_format, extension


@pytest.fixture
def dif_converter_all_packages():
    """All packages in Diff format."""
    file_format = "gcmd_dif"
    extension = "xml"
    return dif_convert_dataset, file_format, extension


@pytest.fixture
def iso_converter_one_package():
    """Single package in ISO format."""
    package_name = "intratrait"
    file_format = "iso19139"
    extension = "xml"
    return iso_convert_dataset, package_name, file_format, extension


@pytest.fixture
def iso_converter_all_packages():
    """All packages in ISO format."""
    file_format = "iso19139"
    extension = "xml"
    return iso_convert_dataset, file_format, extension


@pytest.fixture
def dcat_ap_converter_all_packages():
    """All packages in DCAT-AP format."""
    file_format = "dcat-ap-ch"
    extension = "xml"
    return dcat_ap_convert_dataset, file_format, extension


@pytest.fixture
def ris_converter_one_package():
    """Single package in RIS format."""
    package_name = "bioclim_plus"
    file_format = "ris"
    extension = "ris"
    return ris_convert_dataset, package_name, file_format, extension


@pytest.fixture
def ris_converter_all_packages():
    """All packages in RIS format."""
    file_format = "ris"
    extension = "ris"
    return ris_convert_dataset, file_format, extension


@pytest.fixture
def metadata_keys():
    """List of keys required for an EnviDat metadata Record."""
    return [
        "author",
        "author_email",
        "creator_user_id",
        "date",
        "doi",
        "extras",
        "funding",
        "id",
        "isopen",
        # "language",
        "license_id",
        "license_title",
        "license_url",
        "maintainer",
        "maintainer_email",
        "metadata_created",
        "metadata_modified",
        "name",
        "notes",
        "num_resources",
        "num_tags",
        "organization",
        "owner_org",
        "private",
        "publication",
        "publication_state",
        "related_publications",
        "resource_type",
        "resource_type_general",
        "spatial",
        "spatial_info",
        "state",
        "subtitle",
        "title",
        "type",
        "url",
        "version",
        "resources",
        "tags",
        "groups",
        "relationships_as_subject",
        "relationships_as_object",
    ]


@pytest.fixture
def example_ckan_dict():
    """CKAN metadata dict example for use in tests."""
    return {
        "author": '[{"name": "Müller", "affiliation": "WSL", "affiliation_03": "", "given_name": "Kalin", "identifier": "", "email": "kalin.müller@wsl.ch", "affiliation_02": ""}, {"name": "Fraefel", "affiliation": "WSL", "affiliation_03": "", "given_name": "Marielle", "identifier": "", "email": "marielle.fraefel@wsl.ch", "affiliation_02": ""}, {"name": "Cioldi", "affiliation": "WSL", "affiliation_03": "", "given_name": "Fabrizio", "identifier": "", "email": "fabrizio.cioldi@wsl.ch", "affiliation_02": ""}, {"name": "Camin", "affiliation": "FOEN", "affiliation_03": "", "affiliation_02": "", "identifier": "", "email": "paolo.camin@bafu.admin.ch", "given_name": "Paolo"}, {"name": "Fischer", "affiliation": "WSL", "affiliation_03": "", "affiliation_02": "", "identifier": "", "email": "christoph.fischer@wsl.ch", "given_name": "Christoph"}]',
        "author_email": None,
        "creator_user_id": "334cee1e-6afa-4639-88a2-f980e6ff42c3",
        "date": '[{"date": "2013-01-01", "date_type": "collected", "end_date": ""}]',
        "doi": "10.16904/3",
        "funding": '[{"grant_number": "", "institution": "Federal Office for the Environment FOEN", "institution_url": ""}]',
        "id": "99105534-4a3d-4062-a4f9-69933eab4d37",
        "isopen": False,
        "license_id": "wsl-data",
        "license_title": "WSL Data Policy",
        "license_url": "https://www.wsl.ch/en/about-wsl/programmes-and-initiatives/envidat.html",
        "maintainer": '{"affiliation": "WSL", "email": "fabrizio.cioldi@wsl.ch", "identifier": "", "given_name": "Fabrizio", "name": "Cioldi"}',
        "maintainer_email": None,
        "metadata_created": "2016-10-16T22:24:05.567182",
        "metadata_modified": "2019-10-31T22:59:07.778166",
        "name": "10-16904-3",
        "notes": "In 2013–2014, a survey was conducted in Switzerland to update the Forest Access Roads geo-dataset within the framework of the Swiss National Forest Inventory (NFI). The resulting nationwide dataset contains valuable information on truck-accessible forest roads that can be used to transport wood. The survey involved interviewing staff from the approximately 800 local forest services in Switzerland and recording the data first on paper maps and then in digitized form. The data in the NFI on the forest roads could thus be updated and additional information regarding their trafficability for specific categories of truck included. The information has now been attached to the geometries of the Roads and Tracks of the swissTLM3D (release 2012) of the Federal Office of Topography swisstopo. The resulting data are suitable for statistical analyses and modeling, but further (labour-intensive) validation work would be necessary if they are to be used as a basis for applications requiring more spatial accuracy, such as navigation systems. The data are managed at the Swiss Federal Institute for Forest, Snow and Landscape Research (WSL) and are available for third parties for non-commercial use provided they have purchased a TLM license. \r\n\r\n__Related Publication__: [doi: 10.3188/szf.2016.0136](http://dx.doi.org/10.3188/szf.2016.0136)",
        "num_resources": 1,
        "num_tags": 5,
        "organization": {
            "id": "49192b11-adac-4e68-ad64-be15a4321347",
            "name": "nfi",
            "title": "NFI",
            "type": "organization",
            "description": "The Swiss National Forest Inventory records the current state and the changes of the Swiss forest. The survey obtains data about trees, stands, sample plots and through enquiries at the local forest service.\r\n\r\nThe NFI carried out by the Swiss Federal Institute for Forest, Snow and Landscape Research (WSL) in collaboration with the Forest Division at Federal Office for the Environment (FOEN ). The WSL is responsible for the planning, survey and analysis, as well as the scientific interpretation and publication of the NFI. The political interpretation and implementation is done by the Forest Division.\r\n\r\nThe first survey (LFI1) took place from 1983–85, the second survey followed in 1993–95 and the third inventory was carried out 2004–2006. Since 2009, the continuous survey of the fourth NFI (2009–2017) has been in progress.\r\n\r\nMore information: https://www.lfi.ch/lfi/lfi-en.php",
            "image_url": "https://www.lfi.ch/layout/images/logo.gif",
            "created": "2016-07-05T13:46:14.456454",
            "is_organization": True,
            "approval_status": "approved",
            "state": "active",
        },
        "owner_org": "49192b11-adac-4e68-ad64-be15a4321347",
        "private": False,
        "publication": '{"publisher": "Swiss Federal Institute for Forest, Snow and Landscape WSL /  Federal Office for the Environment FOEN ", "publication_year": "2016"}',
        "publication_state": "published",
        "related_publications": "",
        "resource_type": "Dataset",
        "resource_type_general": "dataset",
        "spatial": '{"type": "Polygon", "coordinates": [[[5.95587, 45.81802],[5.95587, 47.80838],[10.49203, 47.80838],[10.49203, 45.81802],[5.95587, 45.81802]]]}',
        "spatial_info": "Switzerland [45.81802 5.95587 47.80838 10.49203]",
        "state": "active",
        "subtitle": "",
        "title": "Forest Access Roads 2013",
        "type": "dataset",
        "url": None,
        "version": "1",
        "extras": [{"key": "dora_link", "value": ""}],
        "resources": [
            {
                "cache_last_updated": None,
                "cache_url": None,
                "created": "2016-10-17T00:26:11.582643",
                "description": "Article describing the dataset, as well as its acquisition and potential uses.\r\nTo access the data please contact Fabrizio Cioldi.",
                "doi": "10.3188/szf.2016.0136",
                "format": "URL",
                "hash": "",
                "id": "72e5616c-5d4d-4aa7-8702-e8d58c3d6f9f",
                "last_modified": None,
                "metadata_modified": None,
                "mimetype": "application/pdf",
                "mimetype_inner": None,
                "name": "Publication Walderschliessungsstrassen 2013",
                "package_id": "99105534-4a3d-4062-a4f9-69933eab4d37",
                "position": 0,
                "resource_type": None,
                "restricted": '{"level": "public", "allowed_users": ""}',
                "size": None,
                "state": "active",
                "url": "https://www.dora.lib4ri.ch/wsl/islandora/object/wsl:5563",
                "url_type": "",
            }
        ],
        "tags": [
            {
                "display_name": "FOREST ACCESS ROADS",
                "id": "aa4c7d84-f08d-4a8a-aeb0-c2e2adbd5982",
                "name": "FOREST ACCESS ROADS",
                "state": "active",
                "vocabulary_id": None,
            },
            {
                "display_name": "FOREST TRANSPORTATION SYSTEMS",
                "id": "2d33b726-2878-480e-8210-0d82271edde2",
                "name": "FOREST TRANSPORTATION SYSTEMS",
                "state": "active",
                "vocabulary_id": None,
            },
            {
                "display_name": "TIMBER HARVESTING",
                "id": "d677b7e0-748e-49cf-8049-bff0d08644d0",
                "name": "TIMBER HARVESTING",
                "state": "active",
                "vocabulary_id": None,
            },
            {
                "display_name": "TRAFFICABILITY",
                "id": "d9003596-a5d9-4eac-91aa-b31d2743a44a",
                "name": "TRAFFICABILITY",
                "state": "active",
                "vocabulary_id": None,
            },
            {
                "display_name": "TRUCKS",
                "id": "3815dcbd-db09-46cb-80da-ca01ba415af5",
                "name": "TRUCKS",
                "state": "active",
                "vocabulary_id": None,
            },
        ],
        "groups": [],
        "relationships_as_subject": [],
        "relationships_as_object": [],
    }


@pytest.fixture
def example_ckan_json():
    """CKAN metadata JSON example for use in tests."""
    return dedent(
        r"""
        {
        "author": "[{\"name\": \"Müller\", \"affiliation\": \"WSL\", \"affiliation_03\": \"\", \"given_name\": \"Kalin\", \"identifier\": \"\", \"email\": \"kalin.müller@wsl.ch\", \"affiliation_02\": \"\"}, {\"name\": \"Fraefel\", \"affiliation\": \"WSL\", \"affiliation_03\": \"\", \"given_name\": \"Marielle\", \"identifier\": \"\", \"email\": \"marielle.fraefel@wsl.ch\", \"affiliation_02\": \"\"}, {\"name\": \"Cioldi\", \"affiliation\": \"WSL\", \"affiliation_03\": \"\", \"given_name\": \"Fabrizio\", \"identifier\": \"\", \"email\": \"fabrizio.cioldi@wsl.ch\", \"affiliation_02\": \"\"}, {\"name\": \"Camin\", \"affiliation\": \"FOEN\", \"affiliation_03\": \"\", \"affiliation_02\": \"\", \"identifier\": \"\", \"email\": \"paolo.camin@bafu.admin.ch\", \"given_name\": \"Paolo\"}, {\"name\": \"Fischer\", \"affiliation\": \"WSL\", \"affiliation_03\": \"\", \"affiliation_02\": \"\", \"identifier\": \"\", \"email\": \"christoph.fischer@wsl.ch\", \"given_name\": \"Christoph\"}]",
        "author_email": null,
        "creator_user_id": "334cee1e-6afa-4639-88a2-f980e6ff42c3",
        "date": "[{\"date\": \"2013-01-01\", \"date_type\": \"collected\", \"end_date\": \"\"}]",
        "doi": "10.16904/3",
        "funding": "[{\"grant_number\": \"\", \"institution\": \"Federal Office for the Environment FOEN\", \"institution_url\": \"\"}]",
        "id": "99105534-4a3d-4062-a4f9-69933eab4d37",
        "isopen": false,
        "license_id": "wsl-data",
        "license_title": "WSL Data Policy",
        "license_url": "https://www.wsl.ch/en/about-wsl/programmes-and-initiatives/envidat.html",
        "maintainer": "{\"affiliation\": \"WSL\", \"email\": \"fabrizio.cioldi@wsl.ch\", \"identifier\": \"\", \"given_name\": \"Fabrizio\", \"name\": \"Cioldi\"}",
        "maintainer_email": null,
        "metadata_created": "2016-10-16T22:24:05.567182",
        "metadata_modified": "2019-10-31T22:59:07.778166",
        "name": "10-16904-3",
        "notes": "In 2013–2014, a survey was conducted in Switzerland to update the Forest Access Roads geo-dataset within the framework of the Swiss National Forest Inventory (NFI). The resulting nationwide dataset contains valuable information on truck-accessible forest roads that can be used to transport wood. The survey involved interviewing staff from the approximately 800 local forest services in Switzerland and recording the data first on paper maps and then in digitized form. The data in the NFI on the forest roads could thus be updated and additional information regarding their trafficability for specific categories of truck included. The information has now been attached to the geometries of the Roads and Tracks of the swissTLM3D (release 2012) of the Federal Office of Topography swisstopo. The resulting data are suitable for statistical analyses and modeling, but further (labour-intensive) validation work would be necessary if they are to be used as a basis for applications requiring more spatial accuracy, such as navigation systems. The data are managed at the Swiss Federal Institute for Forest, Snow and Landscape Research (WSL) and are available for third parties for non-commercial use provided they have purchased a TLM license. \r\n\r\n__Related Publication__: [doi: 10.3188/szf.2016.0136](http://dx.doi.org/10.3188/szf.2016.0136)",
        "num_resources": 1,
        "num_tags": 5,
        "organization": {
            "id": "49192b11-adac-4e68-ad64-be15a4321347",
            "name": "nfi",
            "title": "NFI",
            "type": "organization",
            "description": "The Swiss National Forest Inventory records the current state and the changes of the Swiss forest. The survey obtains data about trees, stands, sample plots and through enquiries at the local forest service.\r\n\r\nThe NFI carried out by the Swiss Federal Institute for Forest, Snow and Landscape Research (WSL) in collaboration with the Forest Division at Federal Office for the Environment (FOEN ). The WSL is responsible for the planning, survey and analysis, as well as the scientific interpretation and publication of the NFI. The political interpretation and implementation is done by the Forest Division.\r\n\r\nThe first survey (LFI1) took place from 1983–85, the second survey followed in 1993–95 and the third inventory was carried out 2004–2006. Since 2009, the continuous survey of the fourth NFI (2009–2017) has been in progress.\r\n\r\nMore information: https://www.lfi.ch/lfi/lfi-en.php",
            "image_url": "https://www.lfi.ch/layout/images/logo.gif",
            "created": "2016-07-05T13:46:14.456454",
            "is_organization": true,
            "approval_status": "approved",
            "state": "active"
        },
        "owner_org": "49192b11-adac-4e68-ad64-be15a4321347",
        "private": false,
        "publication": "{\"publisher\": \"Swiss Federal Institute for Forest, Snow and Landscape WSL /  Federal Office for the Environment FOEN \", \"publication_year\": \"2016\"}",
        "publication_state": "published",
        "related_publications": "",
        "resource_type": "Dataset",
        "resource_type_general": "dataset",
        "spatial": "{\"type\": \"Polygon\", \"coordinates\": [[[5.95587, 45.81802],[5.95587, 47.80838],[10.49203, 47.80838],[10.49203, 45.81802],[5.95587, 45.81802]]]}",
        "spatial_info": "Switzerland [45.81802 5.95587 47.80838 10.49203]",
        "state": "active",
        "subtitle": "",
        "title": "Forest Access Roads 2013",
        "type": "dataset",
        "url": null,
        "version": "1",
        "extras": [
            {
            "key": "dora_link",
            "value": ""
            }
        ],
        "resources": [
            {
            "cache_last_updated": null,
            "cache_url": null,
            "created": "2016-10-17T00:26:11.582643",
            "description": "Article describing the dataset, as well as its acquisition and potential uses.\r\nTo access the data please contact Fabrizio Cioldi.",
            "doi": "10.3188/szf.2016.0136",
            "format": "URL",
            "hash": "",
            "id": "72e5616c-5d4d-4aa7-8702-e8d58c3d6f9f",
            "last_modified": null,
            "metadata_modified": null,
            "mimetype": "application/pdf",
            "mimetype_inner": null,
            "name": "Publication Walderschliessungsstrassen 2013",
            "package_id": "99105534-4a3d-4062-a4f9-69933eab4d37",
            "position": 0,
            "resource_type": null,
            "restricted": "{\"level\": \"public\", \"allowed_users\": \"\"}",
            "size": null,
            "state": "active",
            "url": "https://www.dora.lib4ri.ch/wsl/islandora/object/wsl:5563",
            "url_type": ""
            }
        ],
        "tags": [
            {
            "display_name": "FOREST ACCESS ROADS",
            "id": "aa4c7d84-f08d-4a8a-aeb0-c2e2adbd5982",
            "name": "FOREST ACCESS ROADS",
            "state": "active",
            "vocabulary_id": null
            },
            {
            "display_name": "FOREST TRANSPORTATION SYSTEMS",
            "id": "2d33b726-2878-480e-8210-0d82271edde2",
            "name": "FOREST TRANSPORTATION SYSTEMS",
            "state": "active",
            "vocabulary_id": null
            },
            {
            "display_name": "TIMBER HARVESTING",
            "id": "d677b7e0-748e-49cf-8049-bff0d08644d0",
            "name": "TIMBER HARVESTING",
            "state": "active",
            "vocabulary_id": null
            },
            {
            "display_name": "TRAFFICABILITY",
            "id": "d9003596-a5d9-4eac-91aa-b31d2743a44a",
            "name": "TRAFFICABILITY",
            "state": "active",
            "vocabulary_id": null
            },
            {
            "display_name": "TRUCKS",
            "id": "3815dcbd-db09-46cb-80da-ca01ba415af5",
            "name": "TRUCKS",
            "state": "active",
            "vocabulary_id": null
            }
        ],
        "groups": [],
        "relationships_as_subject": [],
        "relationships_as_object": []
        }"""
    ).strip()



import base64
import json
from logging import getLogger

import requests

from envidat.converters.datacite_converter import convert_datacite
from envidat.utils import load_dotenv_if_in_debug_mode

log = getLogger(__name__)

# Load config from environment variables
load_dotenv_if_in_debug_mode(".env")


# TODO review if DOIs should continue to be reserved in CKAN database!!!!

# TODO note in documentation for config that environment variables with a hash
#   must be enclosed in quotes

# TODO review where to store environment variables

# TODO review dependencies management with PDM


def reserve_draft_doi_datacite(metadata_record: dict) -> str | None:
    """Reserve a DOI identifer in "Draft" state with DataCite.

       If DOI not available from input metadata_record then new DOI
       generated by DataCite (using default EnviDat DOI suffix in config.)

       For DataCite documentation of this process see:
       https://support.datacite.org/docs/api-create-dois#create-an-identifier-in-draft-state

    Args:
        metadata_record (dict): Individual EnviDat metadata entry record dictionary.

    Returns:
        str|None: DOI reserved in DataCite or None if DOI reservation failed
    """
    # Extract variables from config needed to call DataCite API
    try:
        api_url = os.environ("DATACITE_API_URL")
        client_id = os.environ("DATACITE_CLIENT_ID")
        password = os.envion("DATACITE_PASSWORD")
        doi_prefix = os.environ("DOI_PREFIX")
    except KeyError as e:
        log.error(f"KeyError: {e} does not exist in config")
        return None

    payload = {"data": {"type": "dois"}}

    # If DOI already exists then include it in payload
    doi = metadata_record.get("doi")
    if doi:
        payload["data"]["attributes"] = {"doi": doi}
    # Else instruct DataCite API to auto-generate DOI by only adding
    # "prefix" attribute to payload, this auto-generates a random DOI suffix
    else:
        payload["data"]["attributes"] = {"prefix": doi_prefix}

    # Convert payload to JSON and then send POST request to DataCite API
    payload_json = json.dumps(payload)
    headers = {"Content-Type": "application/vnd.api+json"}

    r = requests.post(
        api_url, headers=headers, auth=(client_id, password), data=payload_json
    )

    # Return DOI
    if r.status_code == 201 or r.status_code == 200:
        reserved_doi = r.json().get("data").get("id")
        if reserved_doi:
            return reserved_doi
        else:
            log.error(
                f"Error cannot parse reserved DOI from DataCite response: "
                f"{r.json()}"
            )
            return None
    else:
        log.error(f"Error reserving DOI on DataCite:  " f"HTTP Code {r.status_code}")
        log.error(f"Error:{r.json()}")
        return None


# TODO investigate not reserving DOI at DataCite and instead directly
#  publish new dataset by reserving DOI within CKAN or other database
def publish_datacite(metadata_record: dict, is_update=False) -> dict | None:
    """Publish a EnviDat record in EnviDat using the "publish" event.

       Converts EnviDat record to DataCite XML format before publication.

       For DataCite documentation of this process see:
       https://support.datacite.org/docs/api-create-dois#changing-the-doi-state
       https://support.datacite.org/docs/api-create-dois#provide-metadata-in-formats
       -other-than-json

    Args:
        metadata_record (dict): Individual EnviDat metadata entry record
                                    dictionary.
        is_update (bool): If true then updates existing DOI,
                          else creates new DOI.
                          Default value is False.

    Returns:
        str/None: DOI reserved in DataCite or None if DOI reservation failed
    """
    # Extract variables from config needed to call DataCite API
    try:
        api_url = os.environ("DATACITE_API_URL")
        client_id = os.enviro("DATACITE_CLIENT_ID")
        password = os.environ("DATACITE_PASSWORD")
        site_url = os.environ("SITE_DATASET_URL")
    except KeyError as e:
        log.error(f"KeyError: {e} does not exist in config")
        return None

    # Get DOI
    doi = metadata_record.get("doi")
    if not doi:
        log.error("ERROR record does not have a 'doi' value")
        return None

    # Get metadata record URL
    name = metadata_record.get("name", metadata_record["id"])
    url = f"{site_url}/{name}"

    # Convert metadata record to DataCite XML and encode to base64 formatted
    # string
    xml = convert_datacite(metadata_record)
    if xml:
        xml_encoded = xml_to_base64(xml)
    else:
        log.error("ERROR unable to convert record to DataCite XML format")
        return None

    # Create payload
    payload = {
        "data": {
            "id": doi,
            "type": "dois",
            "attributes": {
                "event": "publish",
                "doi": doi,
                "url": url,
                "xml": xml_encoded,
            },
        }
    }

    # Convert payload to JSON
    payload_json = json.dumps(payload)
    headers = {"Content-Type": "application/vnd.api+json"}

    # If is_update is True update DOI that is already registered
    if is_update:
        api_url = f"{api_url}/{doi}"
        r = requests.put(
            api_url, headers=headers, auth=(client_id, password), data=payload_json
        )
    # Else create new DOI
    else:
        r = requests.post(
            api_url, headers=headers, auth=(client_id, password), data=payload_json
        )

    # Return DOI if successful, else return error message
    # Status code 201 indicates a new DOI was created
    # Status code 200 indicates an existing DOI was updated
    # TODO implement try/except error handling
    # TODO use get() method on status_code
    if r.status_code == 201 or r.status_code == 200:
        published_doi = r.json().get("data").get("id")
        if published_doi:
            return {"status_code": r.status_code, "result": published_doi}
        else:
            return {
                "status_code": 500,
                "result": f"Failed to parse published DOI from DataCite "
                f"response: {r.json()}",
            }
    else:
        return {
            "status_code": r.status_code,
            "result": f"Failed publishing DOI {doi} on DataCite:  {r.json()}",
        }


def xml_to_base64(xml: str) -> str:
    """Converts XML formatted string to base64 formatted string.

       Returns string in base64 format (not bytes)

    Args:
        xml (str): String in XML format

    Returns:
        str: base64 formatted string conversion of input xml_str
    """
    if isinstance(xml, str):
        xml_bytes = xml.encode("utf-8")
        xml_encoded = base64.b64encode(xml_bytes)
        xml_str = xml_encoded.decode()
        return xml_str



"""Tests for package converters."""

import os
from collections import Counter, OrderedDict

from xmltodict import parse, unparse

from envidat.api.v1 import get_metadata_list_with_resources, get_package
from envidat.converters.dcat_ap_converter import wrap_packages_dcat_ap_xml
from envidat.utils import get_url


def get_ckan_exporter_endpoint(
    package: dict,
    file_format: str,
    extension: str,
    host: str = "https://www.envidat.ch",
) -> str:
    """Get CKAN endpoint used to export datasets in various formats.

    Args:
        package (dict): EnviDat package in dictionary form.
        file_format (str): Format of file used in CKAN endpoint (example: "")
        extension (str): Extension used in CKAN endpoint (example: "xml")
        host (str): Host used in CKAN endpoint (default: "https://www.envidat.ch")

    Returns:
        str: String of CKAN endpoint used to export datasets in format compatible with
        arguments passed.
    """
    if "API_HOST" in os.environ:
        host = os.getenv("API_HOST")

    package_name = package.get("name", "")
    if package_name:
        return f"{host}/dataset/{package_name}/export/{file_format}.{extension}"
    else:
        package_id = package.get("id", "")
        if package_id:
            return f"{host}/dataset/{package_id}/export/{file_format}.{extension}"
        else:
            raise ValueError(
                f"Failed to get CKAN endpoint string for {file_format} format."
            )


def get_converters_one_package(
    convert_dataset, package_name, file_format, extension
) -> tuple[str, str]:
    """Get CKAN output and corresponding converter output for one package.

    Args:
        convert_dataset (function): Function used to convert dataset.
        package_name (str): Name of package.
        file_format (str): Format of file used in CKAN endpoint (example: "")
        extension (str): Extension used in CKAN endpoint (example: "xml")

    Returns:
        tuple (<str: ckan_output>, <str: converter_output>): CKAN output and
        corresponding converter output for one package.
    """
    package = get_package(package_name)

    ckan_endpoint = get_ckan_exporter_endpoint(package, file_format, extension)
    request = get_url(ckan_endpoint)
    ckan_output = request.content.decode()

    converter_output = convert_dataset(package)

    return ckan_output, converter_output


def get_converters_all_packages(
    convert_dataset, file_format, extension
) -> tuple[list, list]:
    """Get CKAN output and corresponding converter output for all packages.

    Args:
        convert_dataset (function): Function used to convert dataset.
        file_format (str): Format of file used in CKAN endpoint (example: "")
        extension (str): Extension used in CKAN endpoint (example: "xml")

    Returns:
        tuple (<list: ckan_packages>, <list: converter_packages>):
        List of all packages converted from CKAN and list of all packages converted
        using convert_dataset function.
    """
    packages = get_metadata_list_with_resources()
    ckan_packages = []
    converter_packages = []

    for package in packages[:10]:

        ckan_endpoint = get_ckan_exporter_endpoint(package, file_format, extension)
        request = get_url(ckan_endpoint)
        ckan_output = request.content.decode()
        if ckan_output.startswith("No converter"):
            # Skip if package conversion fails in ckan
            continue
        ckan_packages.append(ckan_output)

        converter_output = convert_dataset(package)
        converter_packages.append(converter_output)

    return ckan_packages, converter_packages


def get_datacite_converters_one_package(
    convert_dataset, package_name, file_format, extension
) -> tuple[str, str]:
    """Get DacaCite formatted CKAN output and DataCite converter output for one package.

    Args:
        convert_dataset (function): Function used to convert dataset.
        package_name (str): Name of package.
        file_format (str): Format of file used in CKAN endpoint (example: "datacite")
        extension (str): Extension used in CKAN endpoint (example: "xml")

    Returns:
        tuple (<str: ckan_output>, <str: converter_output>): DataCite formatted
            CKAN output and corresponding converter output for one package.
    """
    package = get_package(package_name)

    ckan_endpoint = get_ckan_exporter_endpoint(package, file_format, extension)
    request = get_url(ckan_endpoint)
    ckan_output = request.content.decode()

    converter_output = convert_dataset(package)

    return ckan_output, converter_output


def get_datacite_converters_all_packages(
    convert_dataset, file_format, extension
) -> tuple[list, list]:
    """Get DacaCite CKAN output and DataCite converter output for all packages.

    Args:
        convert_dataset (function): Function used to convert dataset.
        file_format (str): Format of file used in CKAN endpoint (example: "datacite")
        extension (str): Extension used in CKAN endpoint (example: "xml")

    Returns:
        tuple (<list: ckan_packages>, <list: converter_packages>):
            DataCite formatted CKAN output and corresponding converter
            output for all packages.
    """
    packages = get_metadata_list_with_resources()
    ckan_packages = []

    converter_packages = []

    for package in packages:

        ckan_endpoint = get_ckan_exporter_endpoint(package, file_format, extension)
        request = get_url(ckan_endpoint)
        ckan_output = request.content.decode()
        ckan_packages.append(ckan_output)

        converter_output = convert_dataset(package)
        converter_packages.append(converter_output)

    return ckan_packages, converter_packages


def convert_datacite_related_identifier(ckan_output) -> str:
    """Correct typo in EnviDat API Datacite output.

    To make the DataCite converters tests pass it was necessary to simulate
    correcting the typo in the CKAN DataCite converter variable
    'related_datasets_base_url': 'https://www.envidat.ch/#/metadata//'
    (the correct url omits the last slash: 'https://www.envidat.ch/#/metadata/').

    Args:
        ckan_output (str): Output produced from CKAN endpoint.

    Returns:
        str: Output produced from CKAN endpoint with "relatedIdentifier"
            key typo corrected.
    """
    # Convert xml to dict
    ckan_out = parse(ckan_output)

    related_ids = (
        ckan_out.get("resource", {})
        .get("relatedIdentifiers", {})
        .get("relatedIdentifier", {})
    )

    if related_ids:
        related_urls = OrderedDict()
        related_urls["relatedIdentifier"] = []

        if type(related_ids) is list:
            for related_id in related_ids:
                related_url = related_id.get("#text", "")
                if related_url:
                    related_urls["relatedIdentifier"] += get_related_identifier(
                        related_url
                    )

        if type(related_ids) is dict:
            related_url = related_ids.get("#text", "")
            if related_url:
                related_urls["relatedIdentifier"] += get_related_identifier(related_url)

        if len(related_urls["relatedIdentifier"]) > 0:
            ckan_out["resource"]["relatedIdentifiers"] = related_urls

    # Convert dict back to xml
    ckan_xml = unparse(ckan_out, pretty=True)

    return ckan_xml


def get_related_identifier(related_url) -> list:
    """Replace double slash with single slash in EnviDat URL.

    Args:
        related_url (str): URL of CKAN output key "relatedIdentifier"

    Returns:
          list: List with a dictionary in format compatible with
                DataCite "relatedIdentifier" tag.
    """
    related_url = related_url.replace(
        "https://www.envidat.ch/#/metadata//", "https://www.envidat.ch/#/metadata/"
    )
    return [
        {
            "#text": related_url,
            "@relatedIdentifierType": "URL",
            "@relationType": "Cites",
        }
    ]


def convert_dif_values(ckan_output):
    """Correct typo in EnviDat API DIF output.

    To make the DIF converter tests pass it was necessary to simulate correcting the
    typo in the CKAN DIF converter 'Use_Constraints' value:
    "Usage constraintes defined by the license" (the correct spelling is "constraints")
    Also simulates removing extra whitespace produced by the CKAN output for
    the 'Dataset_Creator' value.
    """
    # Convert xml to dict
    ckan_out = parse(ckan_output)

    # Simulate correcting 'Use_Constraints' value
    use_constraints = ckan_out.get("DIF", {}).get("Use_Constraints", "")
    if use_constraints:
        use_constraints = use_constraints.replace("constraintes", "constraints")
        ckan_out["DIF"]["Use_Constraints"] = use_constraints

    # Simulate correcting 'Dataset_Creator' value
    dataset_creator = (
        ckan_out.get("DIF", {}).get("Dataset_Citation", {}).get("Dataset_Creator", "")
    )
    if dataset_creator:
        dataset_creator = dataset_creator.replace("  ", " ").replace(" ,", ",").strip()
        ckan_out["DIF"]["Dataset_Citation"]["Dataset_Creator"] = dataset_creator

    # Convert dict back to xml
    ckan_xml = unparse(ckan_out, pretty=True)

    return ckan_xml


def get_dcat_ap_converters_all_packages(
    convert_dataset,
    file_format,
    extension,
) -> tuple[str, str]:
    """DCAT-AP CKAN and corresponding converter XML formatted strings for all packages.

    Note: As of October 14, 2022, the expected CKAN string should be
            'https://www.envidat.ch/opendata/export/dcat-ap-ch.xml'

    Args:
        convert_dataset (function): Function used to convert dataset.
        file_format (str): Format of file used in CKAN endpoint (example: "dcat-ap")
        extension (str): Extension used in CKAN endpoint (example: "xml")

    Returns:
        tuple (<str: ckan_output>, <str: converter_output>): CKAN output and
        corresponding converter output for one package.
    """
    # FROM CKAN
    ckan_endpoint = f"https://www.envidat.ch/opendata/export/{file_format}.{extension}"
    request = get_url(ckan_endpoint)
    ckan_dcat_str = request.content.decode()

    # FROM CONVERTERS
    packages = get_metadata_list_with_resources()
    converted_packages = []

    for package in packages:
        converter_output = convert_dataset(package)
        converted_packages.append(converter_output)

    converterd_dcat_xml = wrap_packages_dcat_ap_xml(converted_packages)

    return ckan_dcat_str, converterd_dcat_xml


def test_bibtex_converters_one_package(bibtex_converter_one_package):
    """Test Bibtex converter for one package."""
    ckan_output, converter_output = get_converters_one_package(
        *bibtex_converter_one_package
    )

    assert ckan_output == converter_output


def test_bibtex_converters_all_packages(bibtex_converter_all_packages):
    """Test Bibtex converter for all packages."""
    ckan_packages, converter_packages = get_converters_all_packages(
        *bibtex_converter_all_packages
    )

    assert ckan_packages == converter_packages


# def test_datacite_converter_one_package(datacite_converter_one_package):
#     """Test DataCite converter for one package."""
#     ckan_output, converter_output = get_datacite_converters_one_package(
#         *datacite_converter_one_package
#     )

#     # Simulate correct CKAN DataCite converter variable 'related_datasets_base_url'
#     ckan_output = convert_datacite_related_identifier(ckan_output)

#     assert ckan_output == converter_output


# def test_datacite_converters_all_packages(datacite_converter_all_packages):
#     """Test DataCite converter for all packages."""
#     ckan_packages, converter_packages = get_datacite_converters_all_packages(
#         *datacite_converter_all_packages
#     )

#     # Simulate correcting CKAN DataCite converter variable 'related_datasets_base_url'
#     corr_ckan_packages = []
#     for package in ckan_packages:
#         corr_package = convert_datacite_related_identifier(package)
#         corr_ckan_packages.append(corr_package)

#     assert corr_ckan_packages == converter_packages


def test_dif_converters_one_package(dif_converter_one_package):
    """Test DIF converter for one package."""
    ckan_output, converter_output = get_converters_one_package(
        *dif_converter_one_package
    )

    # Simulate correct CKAN DIF values
    ckan_output = convert_dif_values(ckan_output)

    # Convert OrderedDict to xml format
    converted_output_xml = unparse(converter_output, pretty=True)

    assert ckan_output == converted_output_xml


def test_dif_converters_all_packages(dif_converter_all_packages):
    """Test DIF converter for all packages."""
    ckan_packages, converter_packages = get_converters_all_packages(
        *dif_converter_all_packages
    )

    # Simulate correct CKAN DIF values
    corr_ckan_packages = []
    for package in ckan_packages:
        corr_package = convert_dif_values(package)
        corr_ckan_packages.append(corr_package)

    # Convert OrderedDict packages to xml format
    converter_packages_xml = []
    for package in converter_packages:
        package_xml = unparse(package, pretty=True)
        converter_packages_xml.append(package_xml)

    assert corr_ckan_packages == converter_packages_xml


def test_iso_converters_one_package(iso_converter_one_package):
    """Test ISO converter for one package."""
    ckan_output, converter_output = get_converters_one_package(
        *iso_converter_one_package
    )

    # Convert OrderedDict packages to xml format
    converted_output_xml = unparse(converter_output, pretty=True)

    assert ckan_output == converted_output_xml


def test_iso_converters_all_packages(iso_converter_all_packages):
    """Test ISO converter for all packages."""
    ckan_packages, converter_packages = get_converters_all_packages(
        *iso_converter_all_packages
    )

    # Convert OrderedDict packages to xml format
    converter_packages_xml = []
    for package in converter_packages:
        package_xml = unparse(package, pretty=True)
        converter_packages_xml.append(package_xml)

    assert ckan_packages == converter_packages_xml


def test_dcat_ap_converters_all_packages(dcat_ap_converter_all_packages):
    """Test DCAT-AP converter for all packages."""
    ckan_xml, converter_xml = get_dcat_ap_converters_all_packages(
        *dcat_ap_converter_all_packages
    )

    ckan_dict_list = parse(ckan_xml)["rdf:RDF"]["dcat:Catalog"]["dcat:dataset"]
    converter_dict_list = parse(converter_xml)["rdf:RDF"]["dcat:Catalog"][
        "dcat:dataset"
    ]
    # Sort datasets
    ckan_dict_list = sorted(
        ckan_dict_list,
        key=lambda x: x["dcat:Dataset"]["dct:identifier"],
    )
    converter_dict_list = sorted(
        converter_dict_list,
        key=lambda x: x["dcat:Dataset"]["dct:identifier"],
    )
    # Sort keys in datasets
    ckan_dict_list = [
        OrderedDict(sorted(package.items())) for package in ckan_dict_list
    ]
    converter_dict_list = [
        OrderedDict(sorted(package.items())) for package in ckan_dict_list
    ]

    # Check has same number of datasets
    assert len(ckan_dict_list) == len(converter_dict_list)

    # Check a dataset contains same keys
    assert Counter(list(ckan_dict_list[0].keys())) == Counter(
        converter_dict_list[0].keys()
    )

    # Add root wrapper
    ckan_xml = wrap_packages_dcat_ap_xml(ckan_dict_list)
    converter_xml = wrap_packages_dcat_ap_xml(converter_dict_list)

    # Compare XMLs
    assert ckan_xml == converter_xml


def test_ris_converters_one_package(ris_converter_one_package):
    """Test RIS converter for one package."""
    ckan_output, converter_output = get_converters_one_package(
        *ris_converter_one_package
    )

    assert ckan_output == converter_output


def test_ris_converters_all_packages(ris_converter_all_packages):
    """Test RIS converters for all packages."""
    ckan_packages, converter_packages = get_converters_all_packages(
        *ris_converter_all_packages
    )

    assert ckan_packages == converter_packages
